"""Custom exceptions for issues specific to vigorish actions/processes."""


class S3BucketException(Exception):
    """Exception raised when an operation requiring S3 bucket access cannot be performed."""

    def __init__(self, message=None):
        if not message:
            message = "Unable to access S3 bucket, please verify AWS credentials are propertly configured."
        super().__init__(message)


class ScrapedDataException(Exception):
    """Exception raised when data identified by file_type, data_set and url_id cannot be found."""

    def __init__(self, file_type, data_set, url_id):
        message = f"Failed to locate scraped data: URL ID: {url_id} (File Type: {file_type}, Data Set: {data_set})"
        super().__init__(message)


class InvalidSeasonException(Exception):
    """Exception raised when the database does not contain data for the requested season."""

    def __init__(self, year):
        message = f"The database does not contain any data for the MLB {year} Season"
        super().__init__(message)


class UnknownPlayerException(Exception):
    """Exception raised when the database does not contain data for the requested player."""

    def __init__(self, mlb_id):
        message = f"The database does not contain any data for a player with MLB ID: {mlb_id}"
        super().__init__(message)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import List

from dacite import from_dict
from sqlalchemy.engine import RowProxy

from vigorish.types import MetricsDict, RowDict
from vigorish.util.dataclass_helpers import get_field_types


@dataclass
class PitchStatsMetrics:
    year: int = field(repr=False, default=0)
    team_id_bbref: str = field(repr=False, default="")
    opponent_team_id_bbref: str = field(repr=False, default="")
    mlb_id: int = field(repr=False, default=0)
    bbref_id: str = field(repr=False, default="")
    stint_number: int = field(repr=False, default=0)
    total_games: int = field(repr=False, default=0)
    games_as_sp: int = field(repr=False, default=0)
    games_as_rp: int = field(repr=False, default=0)
    wins: int = field(repr=False, default=0)
    losses: int = field(repr=False, default=0)
    saves: int = field(repr=False, default=0)
    innings_pitched: float = 0.0
    total_outs: int = field(repr=False, default=0)
    batters_faced: int = field(repr=False, default=0)
    runs: int = field(repr=False, default=0)
    earned_runs: int = field(repr=False, default=0)
    hits: int = field(repr=False, default=0)
    homeruns: int = field(repr=False, default=0)
    strikeouts: int = field(repr=False, default=0)
    bases_on_balls: int = field(repr=False, default=0)
    era: float = 0.0
    whip: float = 0.0
    k_per_nine: float = field(repr=False, default=0.0)
    bb_per_nine: float = field(repr=False, default=0.0)
    hr_per_nine: float = field(repr=False, default=0.0)
    k_per_bb: float = field(repr=False, default=0.0)
    k_rate: float = field(repr=False, default=0.0)
    bb_rate: float = field(repr=False, default=0.0)
    k_minus_bb: float = field(repr=False, default=0.0)
    hr_per_fb: float = field(repr=False, default=0.0)
    pitch_count: int = field(repr=False, default=0)
    strikes: int = field(repr=False, default=0)
    strikes_contact: int = field(repr=False, default=0)
    strikes_swinging: int = field(repr=False, default=0)
    strikes_looking: int = field(repr=False, default=0)
    ground_balls: int = field(repr=False, default=0)
    fly_balls: int = field(repr=False, default=0)
    line_drives: int = field(repr=False, default=0)
    unknown_type: int = field(repr=False, default=0)
    inherited_runners: int = field(repr=False, default=0)
    inherited_scored: int = field(repr=False, default=0)
    wpa_pitch: float = field(repr=False, default=0.0)
    re24_pitch: float = field(repr=False, default=0.0)

    @classmethod
    def from_query_results(cls, results: List[RowProxy]) -> List[PitchStatsMetrics]:
        return [from_dict(data_class=cls, data=cls._get_pitch_stats(dict(row))) for row in results]

    @classmethod
    def _get_pitch_stats(cls, pitch_stats: RowDict, decimal_precision: int = 3) -> MetricsDict:
        dc_fields = get_field_types(cls)
        return {
            k: round(v, decimal_precision) if dc_fields[k] is float else v
            for k, v in pitch_stats.items()
            if v and k in dc_fields
        }



from bullet import Check, colors


class DataSetCheck(Check):
    """Custom Check element that allows a user to select one or more DataSet values."""

    def __init__(
        self,
        prompt: str = "",
        choices: list = None,
        checked_data_sets: dict = None,
        check: str = "√",
        check_color: str = colors.foreground["default"],
        check_on_switch: str = colors.REVERSE,
        word_color: str = colors.foreground["default"],
        word_on_switch: str = colors.REVERSE,
        background_color: str = colors.background["default"],
        background_on_switch: str = colors.REVERSE,
        pad_right=0,
        indent: int = 0,
        align=0,
        margin: int = 0,
        shift: int = 0,
    ):
        super().__init__(
            prompt,
            choices,
            check,
            check_color,
            check_on_switch,
            word_color,
            word_on_switch,
            background_color,
            background_on_switch,
            pad_right,
            indent,
            align,
            margin,
            shift,
        )
        if checked_data_sets:
            self.checked = [choice in checked_data_sets for choice in choices]



from dataclasses import asdict
from functools import cached_property
from pathlib import Path
from zipfile import ZipFile

from dataclass_csv import DataclassReader
from events import Events

import vigorish.database as db
from vigorish.enums import DataSet
from vigorish.tasks.base import Task
from vigorish.util.result import Result
from vigorish.util.string_helpers import (
    get_bbref_team_id,
    get_date_status_id_from_game_date,
    get_game_date_from_bbref_game_id,
)

DATACLASS_CSV_MAP = {
    db.DateScrapeStatusCsvRow: "scrape_status_date.csv",
    db.GameScrapeStatusCsvRow: "scrape_status_game.csv",
    db.PitchAppScrapeStatusCsvRow: "scrape_status_pitch_app.csv",
    db.BatStatsCsvRow: "bat_stats.csv",
    db.PitchStatsCsvRow: "pitch_stats.csv",
    db.PitchFxCsvRow: "pitchfx.csv",
}

RESTORE_TABLE_ORDER = {
    1: (db.DateScrapeStatusCsvRow, db.DateScrapeStatus),
    2: (db.GameScrapeStatusCsvRow, db.GameScrapeStatus),
    3: (db.PitchAppScrapeStatusCsvRow, db.PitchAppScrapeStatus),
    4: (db.BatStatsCsvRow, db.BatStats),
    5: (db.PitchStatsCsvRow, db.PitchStats),
    6: (db.PitchFxCsvRow, db.PitchFx),
}

BATCH_SIZE = 20000


class RestoreDatabaseTask(Task):
    def __init__(self, app):
        super().__init__(app)
        backup_folder_setting = self.config.all_settings.get("DB_BACKUP_FOLDER_PATH")
        self._team_id_map = {}
        self.backup_folder = backup_folder_setting.current_setting(DataSet.ALL).resolve()
        self.csv_folder = None
        self.csv_map = {}
        self.events = Events(
            (
                "unzip_backup_files_start",
                "unzip_backup_files_complete",
                "restore_database_start",
                "restore_table_start",
                "batch_insert_performed",
                "restore_table_complete",
                "restore_database_complete",
            )
        )

    @cached_property
    def season_id_map(self):
        return db.Season.get_regular_season_map(self.db_session)

    @cached_property
    def player_id_map(self):
        return db.PlayerId.get_player_id_map(self.db_session)

    @cached_property
    def game_id_map(self):
        return db.GameScrapeStatus.get_game_id_map(self.db_session)

    @cached_property
    def pitch_app_id_map(self):
        return db.PitchAppScrapeStatus.get_pitch_app_id_map(self.db_session)

    @property
    def update_relationships_map(self):
        return {
            db.DateScrapeStatusCsvRow: self.update_date_status_relationships,
            db.GameScrapeStatusCsvRow: self.update_game_status_relationships,
            db.PitchAppScrapeStatusCsvRow: self.update_pitch_app_status_relationships,
            db.BatStatsCsvRow: self.update_bat_stats_relationships,
            db.PitchStatsCsvRow: self.update_pitch_stats_relationships,
            db.PitchFxCsvRow: self.update_pitchfx_relationships,
        }

    def get_team_id_map_for_year(self, year):
        team_id_map_for_year = self._team_id_map.get(year)
        if not team_id_map_for_year:
            team_id_map_for_year = db.Team.get_team_id_map_for_year(self.db_session, year)
            self._team_id_map[year] = team_id_map_for_year
        return team_id_map_for_year

    def update_date_status_relationships(self, dataclass):
        date_status_dict = asdict(dataclass)
        game_date = date_status_dict["game_date"]
        date_status_dict["season_id"] = self.season_id_map[game_date.year]
        return date_status_dict

    def update_game_status_relationships(self, dataclass):
        game_status_dict = asdict(dataclass)
        game_date = game_status_dict["game_date"]
        game_status_dict["season_id"] = self.season_id_map[game_date.year]
        game_status_dict["scrape_status_date_id"] = get_date_status_id_from_game_date(game_date)
        return game_status_dict

    def update_pitch_app_status_relationships(self, dataclass):
        pitch_app_dict = asdict(dataclass)
        game_date = get_game_date_from_bbref_game_id(pitch_app_dict["bbref_game_id"])
        pitch_app_dict["pitcher_id"] = self.player_id_map[pitch_app_dict["pitcher_id_mlb"]]
        pitch_app_dict["season_id"] = self.season_id_map[game_date.year]
        pitch_app_dict["scrape_status_date_id"] = get_date_status_id_from_game_date(game_date)
        pitch_app_dict["scrape_status_game_id"] = self.game_id_map[pitch_app_dict["bbref_game_id"]]
        return pitch_app_dict

    def update_player_stats_relationships(self, dataclass):
        player_stats_dict = asdict(dataclass)
        game_date = get_game_date_from_bbref_game_id(player_stats_dict["bbref_game_id"])
        player_stats_dict["player_id"] = self.player_id_map[player_stats_dict["player_id_mlb"]]
        player_stats_dict["player_team_id"] = self.get_team_id_map_for_year(game_date.year)[
            player_stats_dict["player_team_id_bbref"]
        ]
        player_stats_dict["opponent_team_id"] = self.get_team_id_map_for_year(game_date.year)[
            player_stats_dict["opponent_team_id_bbref"]
        ]
        player_stats_dict["season_id"] = self.season_id_map[game_date.year]
        player_stats_dict["date_id"] = get_date_status_id_from_game_date(game_date)
        player_stats_dict["game_status_id"] = self.game_id_map[player_stats_dict["bbref_game_id"]]
        return player_stats_dict

    def update_bat_stats_relationships(self, dataclass):
        return self.update_player_stats_relationships(dataclass)

    def update_pitch_stats_relationships(self, dataclass):
        pitch_stats_dict = self.update_player_stats_relationships(dataclass)
        pitch_stats_dict["pitch_app_db_id"] = self.pitch_app_id_map[pitch_stats_dict["pitch_app_id"]]
        return pitch_stats_dict

    def update_pitchfx_relationships(self, dataclass):
        pfx_dict = asdict(dataclass)
        game_date = get_game_date_from_bbref_game_id(pfx_dict["bbref_game_id"])
        pitcher_team_id_br = get_bbref_team_id(pfx_dict["pitcher_team_id_bb"])
        opponent_team_id_br = get_bbref_team_id(pfx_dict["opponent_team_id_bb"])
        pfx_dict["pitcher_id"] = self.player_id_map[pfx_dict["pitcher_id_mlb"]]
        pfx_dict["batter_id"] = self.player_id_map[pfx_dict["batter_id_mlb"]]
        pfx_dict["team_pitching_id"] = self.get_team_id_map_for_year(game_date.year)[pitcher_team_id_br]
        pfx_dict["team_batting_id"] = self.get_team_id_map_for_year(game_date.year)[opponent_team_id_br]
        pfx_dict["season_id"] = self.season_id_map[game_date.year]
        pfx_dict["date_id"] = get_date_status_id_from_game_date(game_date)
        pfx_dict["game_status_id"] = self.game_id_map[pfx_dict["bbref_game_id"]]
        pfx_dict["pitch_app_db_id"] = self.pitch_app_id_map[pfx_dict["pitch_app_id"]]
        return pfx_dict

    def execute(self, zip_file=None, csv_folder=None):
        if not zip_file:
            result = self.get_most_recent_backup()
            if result.failure:
                return result
            zip_file = result.value
        self.events.unzip_backup_files_start()
        self.csv_map = self.unzip_csv_files(zip_file)
        self.events.unzip_backup_files_complete()
        result = self.app.prepare_database_for_restore(csv_folder)
        if result.failure:
            return result
        self.events.restore_database_start()
        self.restore_database()
        self.remove_csv_files()
        self.events.restore_database_complete()
        return Result.Ok()

    def get_most_recent_backup(self):
        backup_zip_files = list(Path(self.backup_folder).glob("*.zip"))
        if not backup_zip_files:
            return Result.Fail(f"Error! No backups found in folder:\n{self.backup_folder}")
        backup_zip_files.sort(key=lambda x: x.name, reverse=True)
        return Result.Ok(backup_zip_files[0])

    def unzip_csv_files(self, zip_file):
        with ZipFile(zip_file, mode="r") as zip:
            zip.extractall(path=self.backup_folder)
        self.csv_folder = Path(self.backup_folder).joinpath(zip_file.stem)
        return {dataclass: self.csv_folder.joinpath(csv_file) for dataclass, csv_file in DATACLASS_CSV_MAP.items()}

    def restore_database(self):
        for num in sorted(RESTORE_TABLE_ORDER.keys()):
            (dataclass, db_table) = RESTORE_TABLE_ORDER[num]
            csv_file = self.csv_map[dataclass]
            self.events.restore_table_start(db_table)
            self.restore_table_from_csv(csv_file, dataclass, db_table)
            self.events.restore_table_complete(db_table)

    def restore_table_from_csv(self, csv_file, csv_dataclass, db_table):
        with open(csv_file) as csv:
            reader = DataclassReader(csv, csv_dataclass)
            batch = []
            batch_count = 1
            for csv_row in reader:
                obj_dict = self.update_relationships_map[csv_dataclass](csv_row)
                batch.append(obj_dict)
                if len(batch) < BATCH_SIZE:
                    continue
                self.perform_batch_insert(db_table, batch)
                self.events.batch_insert_performed(batch_count)
                batch_count += 1
            if batch:
                self.perform_batch_insert(db_table, batch)
                self.events.batch_insert_performed(batch_count)

    def perform_batch_insert(self, db_table, batch):
        self.db_engine.execute(db_table.__table__.insert(), batch)
        self.db_session.commit()
        batch.clear()

    def remove_csv_files(self):
        for csv_file in self.csv_map.values():
            csv_file.unlink()
        self.csv_folder.rmdir()



"""Decode json dicts of scraped data to custom objects."""
from dacite import from_dict
from dateutil import parser

from vigorish.patch.bbref_boxscores import BBRefBoxscorePatchList, PatchBBRefBoxscorePitchSequence
from vigorish.patch.bbref_games_for_date import (
    BBRefGamesForDatePatchList,
    PatchBBRefGamesForDateGameID,
)
from vigorish.patch.brooks_games_for_date import (
    BrooksGamesForDatePatchList,
    PatchBrooksGamesForDateBBRefGameID,
    PatchBrooksGamesForDateRemoveGame,
)
from vigorish.patch.brooks_pitchfx import (
    BrooksPitchFxPatchList,
    PatchBrooksPitchFxBatterId,
    PatchBrooksPitchFxDeletePitchFx,
    PatchBrooksPitchFxPitcherId,
)
from vigorish.scrape.bbref_boxscores.models.bat_stats import BBRefBatStats
from vigorish.scrape.bbref_boxscores.models.bat_stats_detail import BBRefBatStatsDetail
from vigorish.scrape.bbref_boxscores.models.boxscore import BBRefBoxscore
from vigorish.scrape.bbref_boxscores.models.boxscore_game_meta import BBRefBoxscoreMeta
from vigorish.scrape.bbref_boxscores.models.boxscore_team_data import BBRefBoxscoreTeamData
from vigorish.scrape.bbref_boxscores.models.half_inning import BBRefHalfInning
from vigorish.scrape.bbref_boxscores.models.pbp_event import BBRefPlayByPlayEvent
from vigorish.scrape.bbref_boxscores.models.pbp_other import BBRefPlayByPlayMiscEvent
from vigorish.scrape.bbref_boxscores.models.pbp_substitution import BBRefInGameSubstitution
from vigorish.scrape.bbref_boxscores.models.pitch_stats import BBRefPitchStats
from vigorish.scrape.bbref_boxscores.models.starting_lineup_slot import BBRefStartingLineupSlot
from vigorish.scrape.bbref_boxscores.models.umpire import BBRefUmpire
from vigorish.scrape.bbref_games_for_date.models.game_info import BBRefGameInfo
from vigorish.scrape.bbref_games_for_date.models.games_for_date import BBRefGamesForDate
from vigorish.scrape.brooks_games_for_date.models.game_info import BrooksGameInfo
from vigorish.scrape.brooks_games_for_date.models.games_for_date import BrooksGamesForDate
from vigorish.scrape.brooks_pitch_logs.models.pitch_log import BrooksPitchLog
from vigorish.scrape.brooks_pitch_logs.models.pitch_logs_for_game import BrooksPitchLogsForGame
from vigorish.scrape.brooks_pitchfx.models.pitchfx import BrooksPitchFxData
from vigorish.scrape.brooks_pitchfx.models.pitchfx_log import BrooksPitchFxLog
from vigorish.util.result import Result


def decode_bbref_games_for_date(json_dict):
    """Convert json dictionary to BbrefGamesForDate object."""
    json_dict.pop("__bbref_games_for_date__")
    json_dict["game_date"] = parser.parse(json_dict["game_date_str"])
    games_dict_list = json_dict.pop("games")
    games_for_date = BBRefGamesForDate(**json_dict)
    games_for_date.games = [decode_bbref_game_info(g) for g in games_dict_list]
    return Result.Ok(games_for_date)


def decode_bbref_game_info(json_dict):
    """Convert json dictionary to BBrefGameInfo object."""
    json_dict.pop("__bbref_game_info__")
    return BBRefGameInfo(**json_dict)


def decode_brooks_games_for_date(json_dict):
    """Convert json dictionary to BrooksGamesForDate object."""
    json_dict.pop("__brooks_games_for_date__")
    json_dict["game_date"] = parser.parse(json_dict["game_date_str"])
    games_dict_list = json_dict.pop("games")
    games_for_date = BrooksGamesForDate(**json_dict)
    games_for_date.games = [decode_brooks_game_info(g) for g in games_dict_list]
    return Result.Ok(games_for_date)


def decode_brooks_game_info(json_dict):
    """Convert json dictionary to BrooksGameInfo object."""
    json_dict.pop("__brooks_game_info__")
    return BrooksGameInfo(**json_dict)


def decode_bbref_boxscore(json_dict):
    """Convert json dictionary to BBRefBoxscore object."""
    json_dict.pop("__bbref_boxscore__")
    game_meta_info_dict = json_dict.pop("game_meta_info")
    away_team_dict = json_dict.pop("away_team_data")
    home_team_dict = json_dict.pop("home_team_data")
    innings_dict_list = json_dict.pop("innings_list")
    umpire_dict_list = json_dict.pop("umpires")
    boxscore = BBRefBoxscore(**json_dict)
    boxscore.game_meta_info = decode_bbref_boxscore_meta(game_meta_info_dict)
    boxscore.away_team_data = decode_boxscore_team_data(away_team_dict)
    boxscore.home_team_data = decode_boxscore_team_data(home_team_dict)
    boxscore.innings_list = [decode_bbref_boxscore_half_inning(i) for i in innings_dict_list]
    boxscore.umpires = [BBRefUmpire(**u) for u in umpire_dict_list]
    return Result.Ok(boxscore)


def decode_bbref_boxscore_meta(json_dict):
    json_dict.pop("__bbref_boxscore_meta__")
    return BBRefBoxscoreMeta(**json_dict)


def decode_boxscore_team_data(json_dict):
    json_dict.pop("__bbref_boxscore_team_data__")
    lineup_dict_list = json_dict.pop("starting_lineup")
    bat_stats_dict_list = json_dict.pop("batting_stats")
    pitch_stats_dict_list = json_dict.pop("pitching_stats")
    team_won = json_dict.pop("team_won")
    team_data = BBRefBoxscoreTeamData(**json_dict)
    team_data.team_won = team_won
    team_data.starting_lineup = [BBRefStartingLineupSlot(**lu) for lu in lineup_dict_list]
    team_data.batting_stats = [decode_bbref_bat_stats(bs) for bs in bat_stats_dict_list]
    team_data.pitching_stats = [BBRefPitchStats(**ps) for ps in pitch_stats_dict_list]
    return team_data


def decode_bbref_bat_stats(json_dict):
    details_dict_list = json_dict.pop("details")
    bat_stats = BBRefBatStats(**json_dict)
    bat_stats.details = [BBRefBatStatsDetail(**d) for d in details_dict_list]
    return bat_stats


def decode_bbref_boxscore_half_inning(json_dict):
    json_dict.pop("__bbref_boxscore_half_inning__")
    game_event_dict_list = json_dict.pop("game_events")
    player_sub_dict_list = json_dict.pop("substitutions")
    misc_event_dict_list = json_dict.pop("misc_events")
    inning = BBRefHalfInning(**json_dict)
    inning.game_events = [decode_bbref_playbyplay_event(e) for e in game_event_dict_list]
    inning.substitutions = [decode_bbref_pbp_substitution(s) for s in player_sub_dict_list]
    inning.misc_events = [decode_bbref_pbp_misc_event(m) for m in misc_event_dict_list]
    return inning


def decode_bbref_playbyplay_event(json_dict):
    json_dict.pop("__bbref_pbp_game_event__")
    return BBRefPlayByPlayEvent(**json_dict)


def decode_bbref_pbp_substitution(json_dict):
    json_dict.pop("__bbref_pbp_in_game_substitution__")
    return BBRefInGameSubstitution(**json_dict)


def decode_bbref_pbp_misc_event(json_dict):
    json_dict.pop("__bbref_pbp_misc_event__")
    return BBRefPlayByPlayMiscEvent(**json_dict)


def decode_brooks_pitch_logs_for_game(json_dict):
    json_dict.pop("__brooks_pitch_logs_for_game__")
    pitch_log_dict_list = json_dict.pop("pitch_logs")
    pitch_logs_for_game = BrooksPitchLogsForGame(**json_dict)
    pitch_logs_for_game.pitch_logs = [decode_brooks_pitch_log(pl) for pl in pitch_log_dict_list]
    return Result.Ok(pitch_logs_for_game)


def decode_brooks_pitch_log(json_dict):
    json_dict.pop("__brooks_pitch_log__")
    return BrooksPitchLog(**json_dict)


def decode_brooks_pitchfx_log(json_dict):
    json_dict.pop("__brooks_pitchfx_log__")
    pfx_dict_list = json_dict.pop("pitchfx_log")
    pitchfx_log = BrooksPitchFxLog(**json_dict)
    pitchfx_log.pitchfx_log = [decode_brooks_pitchfx_data(pfx) for pfx in pfx_dict_list]
    return Result.Ok(pitchfx_log)


def decode_brooks_pitchfx_data(json_dict):
    json_dict.pop("__brooks_pitchfx_data__")
    return from_dict(data_class=BrooksPitchFxData, data=json_dict)


def decode_bbref_games_for_date_patch_list(json_dict):
    json_dict.pop("__bbref_games_for_date_patch_list__")
    patch_dict_list = json_dict.pop("patch_list")
    bbref_games_for_date_patch_list = BBRefGamesForDatePatchList(**json_dict)
    bbref_games_for_date_patch_list.patch_list = [decode_patch_bbref_games_for_date_game_id(p) for p in patch_dict_list]
    return Result.Ok(bbref_games_for_date_patch_list)


def decode_patch_bbref_games_for_date_game_id(json_dict):
    json_dict.pop("__patch_bbref_games_for_date_game_id__")
    return from_dict(data_class=PatchBBRefGamesForDateGameID, data=json_dict)


def decode_bbref_boxscore_patch_list(json_dict):
    json_dict.pop("__bbref_boxscore_patch_list__")
    patch_dict_list = json_dict.pop("patch_list")
    boxscore_patch_list = BBRefBoxscorePatchList(**json_dict)
    boxscore_patch_list.patch_list = [decode_patch_bbref_boxscore_pitch_sequence(p) for p in patch_dict_list]
    return Result.Ok(boxscore_patch_list)


def decode_patch_bbref_boxscore_pitch_sequence(json_dict):
    json_dict.pop("__patch_bbref_boxscore_pitch_sequence__")
    return from_dict(data_class=PatchBBRefBoxscorePitchSequence, data=json_dict)


def decode_brooks_games_for_date_patch_list(json_dict):
    json_dict.pop("__brooks_games_for_date_patch_list__")
    patch_dict_list = json_dict.pop("patch_list")
    brooks_games_for_date_patch_list = BrooksGamesForDatePatchList(**json_dict)
    for patch_json in patch_dict_list:
        if "__patch_brooks_games_for_date_bbref_game_id__" in patch_json:
            patch = decode_patch_brooks_games_for_date_bbref_game_id(patch_json)
        elif "__patch_brooks_games_for_date_remove_game__" in patch_json:
            patch = decode_patch_brooks_games_for_date_remove_game(patch_json)
        else:
            return Result.Fail("JSON file is not in the expected format.")
        brooks_games_for_date_patch_list.patch_list.append(patch)
    return Result.Ok(brooks_games_for_date_patch_list)


def decode_patch_brooks_games_for_date_bbref_game_id(json_dict):
    json_dict.pop("__patch_brooks_games_for_date_bbref_game_id__")
    return from_dict(data_class=PatchBrooksGamesForDateBBRefGameID, data=json_dict)


def decode_patch_brooks_games_for_date_remove_game(json_dict):
    json_dict.pop("__patch_brooks_games_for_date_remove_game__")
    return from_dict(data_class=PatchBrooksGamesForDateRemoveGame, data=json_dict)


def decode_brooks_pitchfx_patch_list(json_dict):
    json_dict.pop("__brooks_pitchfx_patch_list__")
    patch_dict_list = json_dict.pop("patch_list")
    pitchfx_patch_list = BrooksPitchFxPatchList(**json_dict)
    for patch_json in patch_dict_list:
        if "__patch_brooks_pitchfx_batter_id__" in patch_json:
            patch = decode_patch_brooks_pitchfx_batter_id(patch_json)
        elif "__patch_brooks_pitchfx_pitcher_id__" in patch_json:
            patch = decode_patch_brooks_pitchfx_pitcher_id(patch_json)
        elif "__patch_brooks_pitchfx_delete_pitchfx__" in patch_json:
            patch = decode_patch_brooks_pitchfx_delete_pitchfx(patch_json)
        else:
            return Result.Fail("JSON file is not in the expected format.")
        pitchfx_patch_list.patch_list.append(patch)
    return Result.Ok(pitchfx_patch_list)


def decode_patch_brooks_pitchfx_batter_id(json_dict):
    json_dict.pop("__patch_brooks_pitchfx_batter_id__")
    return from_dict(data_class=PatchBrooksPitchFxBatterId, data=json_dict)


def decode_patch_brooks_pitchfx_delete_pitchfx(json_dict):
    json_dict.pop("__patch_brooks_pitchfx_delete_pitchfx__")
    return from_dict(data_class=PatchBrooksPitchFxDeletePitchFx, data=json_dict)


def decode_patch_brooks_pitchfx_pitcher_id(json_dict):
    json_dict.pop("__patch_brooks_pitchfx_pitcher_id__")
    return from_dict(data_class=PatchBrooksPitchFxPitcherId, data=json_dict)



from dataclasses import dataclass, field
from datetime import datetime

import vigorish.database as db
from vigorish.enums import DataSet
from vigorish.patch.base import Patch, PatchList
from vigorish.util.dt_format_strings import DATE_ONLY
from vigorish.util.result import Result


@dataclass
class BBRefGamesForDatePatchList(PatchList):
    @property
    def game_date(self):
        return datetime.strptime(self.url_id, DATE_ONLY) if self.url_id else None

    def __post_init__(self):
        self.data_set = DataSet.BBREF_GAMES_FOR_DATE
        self.patch_list_id = "__bbref_games_for_date_patch_list__"

    def apply(self, data, db_session):
        for patch in self.patch_list:
            data = patch.apply(data)
            patch.apply_to_database(data, db_session)
        return Result.Ok(data)


@dataclass
class PatchBBRefGamesForDateGameID(Patch):
    game_date: str = field(repr=False)
    invalid_bbref_game_id: str
    valid_bbref_game_id: str

    def __post_init__(self):
        self.data_set = DataSet.BBREF_GAMES_FOR_DATE
        self.patch_id = "__patch_bbref_games_for_date_game_id__"

    def apply(self, data):
        for game_info in data.games:
            if game_info.bbref_game_id == self.invalid_bbref_game_id:
                game_info.bbref_game_id = self.valid_bbref_game_id
                break
        return data

    def apply_to_database(self, data, db_session):
        game_status = db.GameScrapeStatus.find_by_bbref_game_id(db_session, self.invalid_bbref_game_id)
        if game_status:
            game_status.bbref_game_id = self.valid_bbref_game_id
            db_session.commit()



from collections import defaultdict

import vigorish.database as db
from vigorish.enums import DataSet, VigFile
from vigorish.util.exceptions import ScrapedDataException
from vigorish.util.result import Result


def update_status_combined_data_list(scraped_data, db_session, bbref_game_ids):
    for bbref_game_id in bbref_game_ids:
        combined_data = scraped_data.get_combined_game_data(bbref_game_id)
        if not combined_data:
            raise ScrapedDataException(file_type=VigFile.COMBINED_GAME_DATA, data_set=DataSet.ALL, url_id=bbref_game_id)
        game_status = db.GameScrapeStatus.find_by_bbref_game_id(db_session, bbref_game_id)
        if not game_status:
            error = f"scrape_status_game does not contain an entry for bbref_game_id: {bbref_game_id}"
            return Result.Fail(error)
        game_status.combined_data_success = 1
        result = update_pitch_apps_with_combined_data(db_session, combined_data)
        if result.failure:
            return result
    return Result.Ok()


def update_pitch_apps_with_combined_data(db_session, combined_data):
    pfx_errors = defaultdict(dict)
    combined_success = []
    away_team_pitch_stats = combined_data["away_team_data"]["pitching_stats"]
    home_team_pitch_stats = combined_data["home_team_data"]["pitching_stats"]
    all_pitch_stats = away_team_pitch_stats + home_team_pitch_stats
    for pitch_stats in all_pitch_stats:
        result = update_pitch_appearance_status_records(db_session, pitch_stats)
        if result.failure:
            return result
        pitch_app_status = result.value
        pitchfx_audit = pitch_stats["pitch_app_pitchfx_audit"]
        if pitch_app_status.pitchfx_error:
            at_bat_ids = pitchfx_audit["at_bat_ids_pitchfx_error"]
            pfx_errors["pitchfx_error"][pitch_stats["pitch_app_id"]] = at_bat_ids
        if pitch_app_status.invalid_pitchfx:
            at_bat_ids = pitchfx_audit["at_bat_ids_invalid_pitchfx"]
            pfx_errors["invalid_pitchfx"][pitch_stats["pitch_app_id"]] = at_bat_ids
        if not (pitch_app_status.pitchfx_error or pitch_app_status.invalid_pitchfx):
            combined_success.append(pitch_stats["pitch_app_id"])
    check_for_remaining_pitch_apps_no_pfx_data(db_session, combined_data["bbref_game_id"])
    db_session.commit()
    return Result.Ok({"combined_success": combined_success, "pfx_errors": pfx_errors})


def update_pitch_appearance_status_records(db_session, pitch_stats):
    pitchfx_audit = pitch_stats["pitch_app_pitchfx_audit"]
    result = get_pitch_app_status(db_session, pitch_stats["pitch_app_id"])
    if result.failure:
        return result
    pitch_app_status = result.value
    pitch_app_status.combined_pitchfx_bbref_data = 1
    pitch_app_status.no_pitchfx_data = 1 if pitchfx_audit["pitch_count_pitchfx"] == 0 else 0
    pitch_app_status.batters_faced_bbref = pitchfx_audit["batters_faced_bbref"]
    pitch_app_status.batters_faced_pitchfx = pitchfx_audit["batters_faced_pitchfx"]
    pitch_app_status.pitch_count_bbref = pitchfx_audit["pitch_count_bbref"]
    pitch_app_status.pitch_count_pitchfx_audited = pitchfx_audit["pitch_count_pitchfx"]
    pitch_app_status.total_at_bats_pitchfx_complete = pitchfx_audit["total_at_bats_pitchfx_complete"]
    pitch_app_status.patched_pitchfx_count = pitchfx_audit["patched_pitchfx_count"]
    pitch_app_status.total_at_bats_patched_pitchfx = pitchfx_audit["total_at_bats_patched_pitchfx"]
    pitch_app_status.missing_pitchfx_count = pitchfx_audit["missing_pitchfx_count"]
    pitch_app_status.total_at_bats_missing_pitchfx = pitchfx_audit["total_at_bats_missing_pitchfx"]
    pitch_app_status.removed_pitchfx_count = pitchfx_audit["removed_pitchfx_count"]
    pitch_app_status.total_at_bats_removed_pitchfx = pitchfx_audit["total_at_bats_removed_pitchfx"]
    pitch_app_status.invalid_pitchfx = 1 if pitchfx_audit["invalid_pitchfx"] else 0
    pitch_app_status.invalid_pitchfx_count = pitchfx_audit["invalid_pitchfx_count"]
    pitch_app_status.total_at_bats_invalid_pitchfx = pitchfx_audit["total_at_bats_invalid_pitchfx"]
    pitch_app_status.pitchfx_error = 1 if pitchfx_audit["pitchfx_error"] else 0
    pitch_app_status.total_at_bats_pitchfx_error = pitchfx_audit["total_at_bats_pitchfx_error"]
    return Result.Ok(pitch_app_status)


def get_pitch_app_status(db_session, pitch_app_id):
    pitch_app_status = db.PitchAppScrapeStatus.find_by_pitch_app_id(db_session, pitch_app_id)
    if not pitch_app_status:
        error = f"scrape_status_pitch_app does not contain an entry for pitch_app_id: {pitch_app_id}"
        return Result.Fail(error)
    return Result.Ok(pitch_app_status)


def check_for_remaining_pitch_apps_no_pfx_data(db_session, bbref_game_id):
    pitch_apps_no_pfx_data = (
        db_session.query(db.PitchAppScrapeStatus)
        .filter_by(bbref_game_id=bbref_game_id)
        .filter_by(scraped_pitchfx=1)
        .filter_by(no_pitchfx_data=1)
        .filter_by(combined_pitchfx_bbref_data=0)
        .all()
    )
    for pitch_app_status in pitch_apps_no_pfx_data:
        pitch_app_status.combined_pitchfx_bbref_data = 1



import os
import subprocess
from typing import List

from bullet import Bullet, colors, keyhandler
from bullet.charDef import ARROW_LEFT_KEY, ARROW_RIGHT_KEY
from bullet.cursor import hide as cursor_hidden
from bullet.utils import moveCursorUp

from vigorish.cli.components.util import print_message
from vigorish.cli.components.viewers.display_page import DisplayPage
from vigorish.constants import EMOJIS, MENU_NUMBERS

NAV_PREV = f"{EMOJIS.get('LEFT')}  Previous"
NAV_NEXT = f"Next {EMOJIS.get('RIGHT')}"
NAV_TOTAL_LEN = 43
NAV_PREV_LEN = len(NAV_PREV)
NAV_NEXT_LEN = len(NAV_NEXT)


class PageViewer(Bullet):
    def __init__(
        self,
        pages: List[DisplayPage],
        prompt: str,
        confirm_only: bool = False,
        yes_choice: str = "YES",
        no_choice: str = "NO",
        heading_color: str = None,
        text_color: str = None,
        wrap_text: bool = True,
    ) -> None:
        self.choices_dict = self.get_prompt_choices(confirm_only, yes_choice, no_choice)
        super(PageViewer, self).__init__(
            prompt=prompt,
            choices=list(self.choices_dict.keys()),
            bullet="",
            bullet_color=colors.foreground["default"],
            word_color=colors.foreground["default"],
            word_on_switch=colors.bright(colors.foreground["cyan"]),
            background_color=colors.background["default"],
            background_on_switch=colors.background["default"],
            shift=0,
            indent=0,
            margin=2,
            pad_right=0,
            align=0,
            return_index=False,
        )
        self.pages = pages
        self.confirm_only = confirm_only
        self.heading_color = heading_color
        self.text_color = text_color
        self.wrap_text = wrap_text
        self.page_count = len(pages)
        self.page_index = 0
        self.needs_update = False

    @property
    def current_page(self):
        return self.pages[self.page_index]

    @keyhandler.register(ARROW_LEFT_KEY)
    def prev_page(self):
        if not hasattr(self, "first_page_displayed"):
            return
        if not self.first_page_displayed():
            self.page_index -= 1
            self.needs_update = True

    @keyhandler.register(ARROW_RIGHT_KEY)
    def next_page(self):
        if not hasattr(self, "last_page_displayed"):
            return
        if not self.last_page_displayed():
            self.page_index += 1
            self.needs_update = True

    def first_page_displayed(self):
        return self.page_index == 0

    def last_page_displayed(self):
        return self.page_index == self.page_count - 1

    def get_prompt_choices(self, confirm_only, yes_choice, no_choice):
        yes_no_choices = {
            f"{MENU_NUMBERS.get(1)}  {yes_choice}": True,
            f"{MENU_NUMBERS.get(2)}  {no_choice}": False,
        }
        confirm_choice = {f"{MENU_NUMBERS.get(1)}  OK": True}
        return confirm_choice if confirm_only else yes_no_choices

    def launch(self):
        return self.interactive_mode() if os.environ.get("INTERACTIVE_MODE") == "YES" else self.script_mode()

    def interactive_mode(self):
        while True:
            subprocess.run(["clear"])
            self.current_page.display(
                text_color=self.text_color,
                heading_color=self.heading_color,
                wrap_text=self.wrap_text,
            )
            self.print_page_nav()
            self.needs_update = False
            self.print_prompt()
            if not self.confirm_only:
                self.renderBullets()
            moveCursorUp(len(self.choices) - self.pos)
            with cursor_hidden():
                while True:
                    user_selection = self.handle_input()
                    if self.needs_update:
                        break
                    if user_selection is not None:
                        return self.choices_dict.get(user_selection)

    def script_mode(self):
        subprocess.run(["clear"])
        for page in self.pages:
            page.display(
                text_color=self.text_color,
                heading_color=self.heading_color,
                wrap_text=self.wrap_text,
            )
        return True

    def print_prompt(self):
        prompt = f'\n{" " * self.indent}{self.prompt}\n'
        print_message(prompt, bold=self.confirm_only)

    def print_page_nav(self):
        if self.page_count <= 1:
            return
        page_number = f"({self.page_index + 1}/{self.page_count})"
        (pad_left, pad_right) = self.get_pad_lengths(len(page_number))
        nav_prev = NAV_PREV if not self.first_page_displayed() else f'{" " * NAV_PREV_LEN}'
        nav_next = NAV_NEXT if not self.last_page_displayed() else f'{" " * NAV_NEXT_LEN}'
        page_nav = f'\n{nav_prev}{" " * pad_left}{page_number}{" " * pad_right}{nav_next}'
        print_message(page_nav, fg=self.text_color)

    def get_pad_lengths(self, page_number_len):
        remaining = NAV_TOTAL_LEN - NAV_PREV_LEN - NAV_NEXT_LEN - page_number_len
        pad = int(remaining / 2)
        return (pad, pad + 1) if remaining % 2 else (pad, pad)



from dataclasses import dataclass, field

from vigorish.enums import DataSet
from vigorish.patch.base import Patch, PatchList
from vigorish.util.result import Result
from vigorish.util.string_helpers import validate_bbref_game_id


@dataclass
class BBRefBoxscorePatchList(PatchList):
    @property
    def game_date(self):
        result = validate_bbref_game_id(self.url_id)
        return result.value["game_date"] if result.success else None

    def __post_init__(self):
        self.data_set = DataSet.BBREF_BOXSCORES
        self.patch_list_id = "__bbref_boxscore_patch_list__"


@dataclass
class PatchBBRefBoxscorePitchSequence(Patch):
    bbref_game_id: str
    inning_id: str
    pbp_table_row_number: int = field(repr=False)
    old_pitch_sequence: str = field(repr=False)
    new_pitch_sequence: str = field(repr=False)

    def __post_init__(self):
        self.data_set = DataSet.BBREF_BOXSCORES
        self.patch_id = "__patch_bbref_boxscore_pitch_sequence__"

    def apply(self, data):
        inning_matches = [inning for inning in data.innings_list if inning.inning_id == self.inning_id]
        if not inning_matches:
            error = f"Unable to locate the inning identified in this patch: {self.inning_id}"
            return Result.Fail(error)
        if len(inning_matches) > 1:
            error = "More than one inning was found that matches the inning identified in this patch: {self.inning_id}"
            return Result.Fail(error)
        inning = inning_matches[0]
        event_matches = [
            event for event in inning.game_events if event.pbp_table_row_number == self.pbp_table_row_number
        ]
        if not event_matches:
            error = (
                "Unable to locate the game event identified by pbp_table_row_number in this "
                f"patch: {self.pbp_table_row_number}"
            )
            return Result.Fail(error)
        if len(event_matches) > 1:
            error = (
                "More than one game event was found that matches the pbp_table_row_number "
                f"identified in this patch: {self.pbp_table_row_number}"
            )
            return Result.Fail(error)
        event_matches[0].pitch_sequence = self.new_pitch_sequence
        return Result.Ok(data)



"""Menu item that returns the user to the previous menu."""
from vigorish.cli.menu_item import MenuItem
from vigorish.constants import EMOJIS
from vigorish.util.result import Result


class ReturnToParent(MenuItem):
    def __init__(self, app, menu_item_text):
        super().__init__(app)
        self.menu_item_text = menu_item_text
        self.menu_item_emoji = EMOJIS.get("BACK", "")
        self.exit_menu = True

    def launch(self):
        return Result.Ok(self.exit_menu)



from dataclasses import dataclass


@dataclass
class BBRefBatStatsDetail:
    """Additional details for important batting events."""

    count: str = ""
    stat: str = ""

    def as_dict(self):
        """Convert detailed batting stats to a dictionary."""
        return {"count": int(self.count), "stat": self.stat}



"""Scrape MLB player data"""
from datetime import date, datetime, timedelta
from json.decoder import JSONDecodeError

from events import Events
from sqlalchemy.exc import DBAPIError, SQLAlchemyError

import vigorish.database as db
from vigorish.tasks.base import Task
from vigorish.util.request_url import request_url_with_retries
from vigorish.util.result import Result
from vigorish.util.string_helpers import fuzzy_match

MLB_PLAYER_SEARCH_URL = "http://lookup-service-prod.mlb.com/json/named.search_player_all.bam"
MLB_DATE_FORMAT = "%Y-%m-%dT%H:%M:%S"


class ScrapeMlbPlayerInfoTask(Task):
    def __init__(self, app):
        super().__init__(app)
        self.events = Events(
            (
                "error_occurred",
                "scrape_player_info_start",
                "scrape_player_info_complete",
            )
        )

    def execute(self, name, bbref_id, game_date):
        return (
            self.get_search_url(name)
            .on_success(request_url_with_retries)
            .on_success(self.decode_json_response)
            .on_success(self.get_player_data, name, game_date)
            .on_success(self.parse_player_data, bbref_id)
            .on_success(self.add_player_to_database)
        )

    def get_search_url(self, name):
        self.events.scrape_player_info_start(name)
        split = name.split()
        if not split or len(split) <= 1:  # pragma: no cover
            return Result.Fail(f"Name was not in an expected format: {name}")
        name_part = ("%20".join(split[1:])).upper()
        url = f"{MLB_PLAYER_SEARCH_URL}?sport_code='mlb'&name_part='{name_part}%25'&active_sw='Y'"
        return Result.Ok(url)

    def decode_json_response(self, response):
        query_results = ""
        try:
            resp_json = response.json()
            query_results = resp_json["search_player_all"]["queryResults"]
            num_results = int(query_results["totalSize"])
            return Result.Ok((query_results, num_results))
        except (JSONDecodeError, KeyError) as e:  # pragma: no cover
            error = f"Failed to decode HTTP response as JSON: {repr(e)}\n{e.response.text}"
            return Result.Fail(error)
        except ValueError:  # pragma: no cover
            error = f"Failed to parse number of results from search response: {query_results}"
            return Result.Fail(error)

    def get_player_data(self, results_tuple, name, game_date):
        results = results_tuple[0]
        num_results = results_tuple[1]
        if num_results > 1:
            player_list = results["row"]
            player_data = self.find_best_match(player_list, name, game_date)
        else:
            player_data = results["row"]
        return Result.Ok(player_data)

    def find_best_match(self, player_list, name, game_date):
        player_id_name_map = {player["player_id"]: player["name_display_first_last"] for player in player_list}
        player_info_dict = {player["player_id"]: player for player in player_list}
        possible_matches = fuzzy_match(name, player_id_name_map)
        if len(possible_matches) == 1:
            return player_info_dict[possible_matches[0]["result"]]
        return self.compare_mlb_debut(possible_matches, player_info_dict, game_date)

    def compare_mlb_debut(self, possible_matches, player_info_dict, game_date):
        info_dict_list = [player_info_dict[possible_match["result"]] for possible_match in possible_matches]
        for player_info in info_dict_list:
            if not player_info["pro_debut_date"]:  # pragma: no cover
                player_info["since_debut"] = timedelta.max.days
                continue
            player_mlb_debut = datetime.strptime(player_info["pro_debut_date"], MLB_DATE_FORMAT)
            player_info["since_debut"] = abs((game_date - player_mlb_debut.date()).days)
        info_dict_list.sort(key=lambda x: x["since_debut"])
        return info_dict_list[0]

    def parse_player_data(self, player_data, bbref_id):
        try:
            feet = int(player_data["height_feet"])
            inches = int(player_data["height_inches"])
            height_total_inches = (feet * 12) + inches
        except ValueError:  # pragma: no cover
            height_total_inches = 0

        try:
            debut = datetime.strptime(player_data["pro_debut_date"], MLB_DATE_FORMAT).date()
            birth_date = datetime.strptime(player_data["birth_date"], MLB_DATE_FORMAT).date()
        except ValueError:  # pragma: no cover
            debut = date.min
            birth_date = date.min

        player_dict = {
            "name_first": player_data["name_first"],
            "name_last": player_data["name_last"],
            "name_given": player_data["name_first"],
            "bats": player_data["bats"],
            "throws": player_data["throws"],
            "weight": player_data["weight"],
            "height": height_total_inches,
            "debut": debut,
            "birth_year": birth_date.year,
            "birth_month": birth_date.month,
            "birth_day": birth_date.day,
            "birth_country": player_data["birth_country"],
            "birth_state": player_data["birth_state"],
            "birth_city": player_data["birth_city"],
            "bbref_id": bbref_id,
            "mlb_id": player_data["player_id"],
            "missing_mlb_id": False,
        }
        return Result.Ok(player_dict)

    def add_player_to_database(self, player_dict):
        try:
            new_player = db.Player(**player_dict)
            self.db_session.add(new_player)
            self.db_session.commit()
            self.events.scrape_player_info_complete(new_player)
            return Result.Ok(new_player)
        except (SQLAlchemyError, DBAPIError) as e:  # pragma: no cover
            return Result.Fail(f"Error: {repr(e)}")



"""Reusable menu prompts to get various values/data types from the user."""
import subprocess
from typing import List, Optional

from bullet import Check, colors
from getch import pause

import vigorish.database as db
from vigorish.cli.components.data_set_check import DataSetCheck
from vigorish.cli.components.date_input import DateInput
from vigorish.cli.components.util import (
    create_bullet_prompt,
    create_scrolling_prompt,
    print_error,
    print_heading,
    print_message,
)
from vigorish.constants import DATA_SET_FROM_NAME_MAP, DATA_SET_TO_NAME_MAP, EMOJIS, MENU_NUMBERS
from vigorish.enums import DataSet, VigFile
from vigorish.util.result import Result
from vigorish.util.string_helpers import wrap_text


def yes_no_prompt(prompt, wrap=True, max_line_len=70):
    if wrap:
        prompt = wrap_text(prompt, max_len=max_line_len)
    choices = {
        f"{MENU_NUMBERS.get(1)}  YES": True,
        f"{MENU_NUMBERS.get(2)}  NO": False,
    }
    selected = create_bullet_prompt(prompt, list(choices.keys()), wrap, max_line_len).launch()
    return choices.get(selected)


def yes_no_cancel_prompt(prompt, wrap=True, max_line_len=70):
    if wrap:
        prompt = wrap_text(prompt, max_len=max_line_len)
    choices = {
        f"{MENU_NUMBERS.get(1)}  YES": True,
        f"{MENU_NUMBERS.get(2)}  NO": False,
        f"{MENU_NUMBERS.get(3)}  CANCEL": None,
    }
    selected = create_bullet_prompt(prompt, list(choices.keys()), wrap, max_line_len).launch()
    choice_value = choices.get(selected)
    return Result.Fail("") if "CANCEL" in selected else Result.Ok(choice_value)


def single_date_prompt(prompt):
    user_date = None
    while not user_date:
        subprocess.run(["clear"])
        date_prompt = DateInput(prompt=prompt)
        result = date_prompt.launch()
        if result:
            user_date = result
    return user_date


def user_options_prompt(
    choice_map, prompt, wrap=True, clear_screen=True, auto_scroll=True, first_item_is_exit_menu=False
):
    choices = list(choice_map.keys())
    if first_item_is_exit_menu:
        choices.insert(0, f"{EMOJIS.get('BACK')} Return to Previous Menu")
    options_prompt = (
        create_scrolling_prompt(prompt, choices, wrap, max_line_len=70)
        if auto_scroll and len(choices) > 8
        else create_bullet_prompt(prompt, choices, wrap, max_line_len=70)
    )
    if clear_screen:
        subprocess.run(["clear"])
        print()
    selection = options_prompt.launch()
    user_choice = choice_map.get(selection)
    return Result.Ok(user_choice) if user_choice else Result.Fail("")


def season_prompt(db_session, prompt=None, clear_screen=True):
    if not prompt:
        prompt = "Please select a MLB Season from the list below:"
    choices = {
        f"{MENU_NUMBERS.get(num)}  {season.year}": season
        for num, season in enumerate(db.Season.get_all_regular_seasons(db_session), start=1)
    }
    choices[f"{EMOJIS.get('BACK')} Return to Previous Menu"] = None
    return user_options_prompt(choices, prompt=prompt, clear_screen=clear_screen)


def audit_report_season_prompt(audit_report, prompt=None):
    if not prompt:
        prompt = "Select an MLB season from the list below:"
    choices = {f"{MENU_NUMBERS.get(num)}  {year}": year for num, year in enumerate(audit_report.keys(), start=1)}
    choices[f"{EMOJIS.get('BACK')} Return to Previous Menu"] = None
    return user_options_prompt(choices, prompt)


def data_sets_prompt(
    heading: str = None,
    prompt: str = None,
    valid_data_sets: Optional[List[DataSet]] = None,
    checked_data_sets: Optional[List[DataSet]] = None,
):
    if not prompt:
        prompt = "Select one or multiple data sets from the list below:"
    if not valid_data_sets:
        valid_data_sets = [DataSet.ALL]
    instructions = "(use SPACE BAR to select each data set, ENTER to confirm your selections)"
    valid_data_sets_int = sum(int(ds) for ds in valid_data_sets)
    choices = {f"{DATA_SET_TO_NAME_MAP[ds]}": ds for ds in DataSet if valid_data_sets_int & ds == ds}
    if checked_data_sets:
        checked_int = sum(int(ds) for ds in checked_data_sets)
        checked_data_sets = [f"{DATA_SET_TO_NAME_MAP[ds]}" for ds in DataSet if checked_int & ds == ds]
    ds_prompt = DataSetCheck(
        prompt=instructions,
        choices=list(choices.keys()),
        checked_data_sets=checked_data_sets,
        check=EMOJIS.get("CHECK", ""),
        shift=1,
        indent=0,
        margin=2,
        check_color=colors.foreground["default"],
        check_on_switch=colors.foreground["default"],
        background_color=colors.foreground["default"],
        background_on_switch=colors.foreground["default"],
        word_color=colors.foreground["default"],
        word_on_switch=colors.bright(colors.foreground["cyan"]),
    )
    data_sets = []
    while not data_sets:
        subprocess.run(["clear"])
        if heading:
            print_heading(heading, fg="bright_yellow")
        print_message(prompt, wrap=True)
        result = ds_prompt.launch()
        if not result:
            print_error("\nYou must select at least one data set!")
            pause(message="Press any key to continue...")
            continue
        data_sets = [DATA_SET_FROM_NAME_MAP[sel] for sel in result]
    return data_sets


def file_types_prompt(prompt, valid_file_types=VigFile.ALL):
    if not prompt:
        prompt = "Select one or multiple file types from the list below:"
    choices = {f"{f}": f for f in VigFile if int(f) & valid_file_types == f}
    instructions = "(use SPACE BAR to select each file type, ENTER to confirm your selections)"
    file_types_prompt = Check(
        prompt=instructions,
        choices=list(choices.keys()),
        check=EMOJIS.get("CHECK", ""),
        shift=1,
        indent=0,
        margin=2,
        check_color=colors.foreground["default"],
        check_on_switch=colors.foreground["default"],
        word_color=colors.foreground["default"],
        word_on_switch=colors.bright(colors.foreground["cyan"]),
        background_color=colors.background["default"],
        background_on_switch=colors.background["default"],
    )
    file_types = []
    while not file_types:
        subprocess.run(["clear"])
        print_message(prompt, fg="bright_yellow", bold=True, underline=True)
        result = file_types_prompt.launch()
        if not result:
            print_error("\nYou must select at least one file type!")
            pause(message="Press any key to continue...")
            continue
        file_types = [choices[sel] for sel in result]
    return file_types


def select_game_prompt(game_ids, prompt=None, use_numbers=True, clear_screen=True):
    if not prompt:
        prompt = "Select a game from the list below:"
    choices = {f"{_get_menu_item_emoji(use_numbers, n)}  {gid}": gid for n, gid in enumerate(game_ids, start=1)}
    choices[f"{EMOJIS.get('BACK')} Return to Main Menu"] = None
    return user_options_prompt(choices, prompt, clear_screen=clear_screen, first_item_is_exit_menu=True)


def _get_menu_item_emoji(use_numbers, num=None):
    if use_numbers and num:
        return MENU_NUMBERS.get(num, str(num))
    return EMOJIS.get("BLUE_DIAMOND")



from dataclasses import dataclass

from vigorish.enums import PlayByPlayEvent


@dataclass
class BBRefInGameSubstitution:
    """Information describing an in-game player substitution."""

    inning_id: str = ""
    inning_label: str = ""
    pbp_table_row_number: str = "0"
    sub_description: str = ""
    incoming_player_name: str = ""
    outgoing_player_name: str = ""
    incoming_player_id_br: str = ""
    outgoing_player_id_br: str = ""
    incoming_player_pos: str = ""
    outgoing_player_pos: str = ""
    lineup_slot: str = "0"
    sub_type: str = ""
    team_id: str = ""

    @property
    def event_type(self):
        return PlayByPlayEvent.SUBSTITUTION

    def as_dict(self):
        return {
            "__bbref_pbp_in_game_substitution__": True,
            "inning_id": self.inning_id,
            "inning_label": self.inning_label,
            "pbp_table_row_number": int(self.pbp_table_row_number),
            "sub_type": self.sub_type,
            "team_id": self.team_id,
            "sub_description": self.sub_description,
            "incoming_player_id_br": self.incoming_player_id_br,
            "incoming_player_pos": self.incoming_player_pos,
            "outgoing_player_id_br": self.outgoing_player_id_br,
            "outgoing_player_pos": self.outgoing_player_pos,
            "lineup_slot": int(self.lineup_slot),
        }



from __future__ import annotations

from dataclasses import dataclass, field
from functools import cached_property
from typing import Dict, List, Union

from dacite import from_dict
from sqlalchemy.engine import RowProxy

from vigorish.data.metrics.constants import PFX_BATTED_BALL_METRICS, PFX_PLATE_DISCIPLINE_METRICS
from vigorish.enums import PitchType
from vigorish.types import RowDict
from vigorish.util.dataclass_helpers import get_field_types
from vigorish.util.string_helpers import format_decimal_bat_stat

PitchFxMetricsDict = Dict[str, Union[bool, int, float, str, PitchType]]
PitchTypeMetricsDict = Dict[PitchType, PitchFxMetricsDict]


@dataclass
class PitchFxMetrics:
    mlb_id: int = field(repr=False, default=0)
    p_throws: str = field(repr=False, default="")
    pitch_type: PitchType = PitchType.NONE
    total_pitches: int = field(repr=False, default=0)
    total_pa: int = field(repr=False, default=0)
    total_at_bats: int = field(repr=False, default=0)
    total_outs: int = field(repr=False, default=0)
    total_hits: int = field(repr=False, default=0)
    total_bb: int = field(repr=False, default=0)
    total_k: int = field(repr=False, default=0)
    avg_speed: float = 0.0
    avg: float = field(repr=False, default=0.0)
    obp: float = field(repr=False, default=0.0)
    slg: float = field(repr=False, default=0.0)
    ops: float = field(repr=False, default=0.0)
    iso: float = field(repr=False, default=0.0)
    fly_ball_rate: float = field(repr=False, default=0.0)
    ground_ball_rate: float = field(repr=False, default=0.0)
    line_drive_rate: float = field(repr=False, default=0.0)
    pop_up_rate: float = field(repr=False, default=0.0)
    bb_rate: float = field(repr=False, default=0.0)
    k_rate: float = field(repr=False, default=0.0)
    hr_per_fb: float = field(repr=False, default=0.0)
    avg_pfx_x: float = field(repr=False, default=0.0)
    avg_pfx_z: float = field(repr=False, default=0.0)
    avg_px: float = field(repr=False, default=0.0)
    avg_pz: float = field(repr=False, default=0.0)
    zone_rate: float = field(repr=False, default=0.0)
    called_strike_rate: float = field(repr=False, default=0.0)
    swinging_strike_rate: float = field(repr=False, default=0.0)
    whiff_rate: float = field(repr=False, default=0.0)
    csw_rate: float = field(repr=False, default=0.0)
    o_swing_rate: float = field(repr=False, default=0.0)
    z_swing_rate: float = field(repr=False, default=0.0)
    swing_rate: float = field(repr=False, default=0.0)
    o_contact_rate: float = field(repr=False, default=0.0)
    z_contact_rate: float = field(repr=False, default=0.0)
    contact_rate: float = field(repr=False, default=0.0)
    total_swings: int = field(repr=False, default=0)
    total_swings_made_contact: int = field(repr=False, default=0)
    total_called_strikes: int = field(repr=False, default=0)
    total_swinging_strikes: int = field(repr=False, default=0)
    total_inside_strike_zone: int = field(repr=False, default=0)
    total_outside_strike_zone: int = field(repr=False, default=0)
    total_swings_inside_zone: int = field(repr=False, default=0)
    total_swings_outside_zone: int = field(repr=False, default=0)
    total_contact_inside_zone: int = field(repr=False, default=0)
    total_contact_outside_zone: int = field(repr=False, default=0)
    total_batted_balls: int = field(repr=False, default=0)
    total_ground_balls: int = field(repr=False, default=0)
    total_line_drives: int = field(repr=False, default=0)
    total_fly_balls: int = field(repr=False, default=0)
    total_pop_ups: int = field(repr=False, default=0)
    total_singles: int = field(repr=False, default=0)
    total_doubles: int = field(repr=False, default=0)
    total_triples: int = field(repr=False, default=0)
    total_homeruns: int = field(repr=False, default=0)
    total_ibb: int = field(repr=False, default=0)
    total_hbp: int = field(repr=False, default=0)
    total_errors: int = field(repr=False, default=0)
    total_sac_hit: int = field(repr=False, default=0)
    total_sac_fly: int = field(repr=False, default=0)
    pitch_type_int: int = field(repr=False, default=0)
    percent: float = 0.0

    @property
    def pitch_name(self) -> str:
        return self.pitch_type.print_name

    @property
    def triple_slash(self) -> str:
        return (
            f"{format_decimal_bat_stat(self.avg)}/"
            f"{format_decimal_bat_stat(self.obp)}/"
            f"{format_decimal_bat_stat(self.slg)} ({format_decimal_bat_stat(self.ops)})"
        )

    def get_bat_stats(self, include_pa_count: bool = False) -> Dict[str, str]:
        total_pa = f" ({self.total_pa})" if include_pa_count else ""
        return {
            "pitch_type": f"{self.pitch_name}{total_pa}",
            "AVG/OBP/SLG (OPS)": self.triple_slash,
            "K%": f"{self.k_rate:.0%}",
            "BB%": f"{self.bb_rate:.0%}",
            "HR/FB": f"{self.hr_per_fb:.0%}",
        }

    def get_plate_discipline_stats(self, include_pitch_count: bool = False) -> Dict[str, str]:
        total_pitches = f" ({self.total_pitches})" if include_pitch_count else ""
        pd_stats = {"pitch_type": f"{self.pitch_name}{total_pitches}"}
        for metric, (rate, total) in PFX_PLATE_DISCIPLINE_METRICS.items():
            pitch_count = f" ({getattr(self, total)})" if include_pitch_count else ""
            pd_stats[metric] = f"{getattr(self, rate):.0%}{pitch_count}"
        return pd_stats

    def get_batted_ball_stats(self, include_bip_count: bool = False) -> Dict[str, str]:
        total_bip = f" ({self.total_batted_balls})" if include_bip_count else ""
        bb_stats = {"pitch_type": f"{self.pitch_name}{total_bip}"}
        for metric, (rate, total) in PFX_BATTED_BALL_METRICS.items():
            bip_count = f" ({getattr(self, total)})" if include_bip_count else ""
            bb_stats[metric] = f"{getattr(self, rate):.0%}{bip_count}"
        return bb_stats

    def get_usage_stats(self, include_pitch_count: bool = False) -> str:
        pitch_count = f" ({self.total_pitches})" if include_pitch_count else ""
        return f"{self.percent:.0%}{pitch_count} {self.avg_speed:.1f}mph"

    @classmethod
    def from_query_results(cls, results: List[RowProxy]) -> List[PitchFxMetrics]:
        return [from_dict(data_class=cls, data=cls._get_pitchfx_metrics(dict(row))) for row in results]

    @classmethod
    def _get_pitchfx_metrics(cls, pfx_metrics: RowDict, decimal_precision: int = 3) -> PitchFxMetricsDict:
        dc_fields = get_field_types(cls)
        pitchfx_metrics_dict = {
            k: round(v, decimal_precision) if dc_fields[k] is float else bool(v) if dc_fields[k] is bool else v
            for k, v in pfx_metrics.items()
            if v and k in dc_fields
        }
        pitchfx_metrics_dict["pitch_type"] = (
            PitchType(pfx_metrics["pitch_type_int"])
            if "pitch_type_int" in pfx_metrics
            else PitchType.from_abbrev(pfx_metrics["pitch_type"])
            if "pitch_type" in pfx_metrics
            else None
        )
        return pitchfx_metrics_dict


@dataclass
class PitchFxMetricsCollection(PitchFxMetrics):
    pitch_type_metrics: Dict[PitchType, PitchFxMetrics] = field(default_factory=dict)

    @property
    def pitch_name(self) -> str:
        return PitchType.ALL.print_name

    @property
    def pitch_types(self) -> List[PitchType]:
        return list(self.pitch_type_metrics.keys())

    @cached_property
    def pitch_mix(self) -> List[Dict[str, str]]:
        return [
            {"pitch_type": pitch_type.print_name, "usage": metrics.get_usage_stats()}
            for pitch_type, metrics in self.pitch_type_metrics.items()
        ]

    def get_bat_stats_pitch_type_splits(self, include_pa_count: bool = False) -> List[Dict[str, str]]:
        bat_stats_pt_splits = [metrics.get_bat_stats(include_pa_count) for metrics in self.pitch_type_metrics.values()]
        bat_stats_pt_splits.insert(0, self.get_bat_stats(include_pa_count))
        return bat_stats_pt_splits

    def get_plate_discipline_pitch_type_splits(self, include_pitch_count: bool = False) -> List[Dict[str, str]]:
        pd_split_stats = [
            metrics.get_plate_discipline_stats(include_pitch_count) for metrics in self.pitch_type_metrics.values()
        ]
        pd_split_stats.insert(0, self.get_plate_discipline_stats(include_pitch_count))
        return pd_split_stats

    def get_batted_ball_pitch_type_splits(self, include_bip_count: bool = False) -> List[Dict[str, str]]:
        bb_split_stats = [
            metrics.get_batted_ball_stats(include_bip_count) for metrics in self.pitch_type_metrics.values()
        ]
        bb_split_stats.insert(0, self.get_batted_ball_stats(include_bip_count))
        return bb_split_stats

    def get_usage_stats_for_pitch_type(self, pitch_type: PitchType, include_pitch_count: bool = False) -> str:
        return (
            self.pitch_type_metrics[pitch_type].get_usage_stats(include_pitch_count)
            if pitch_type in self.pitch_types
            else "0% (N/A)"
        )

    @classmethod
    def from_query_results(cls, all_pt_results: RowDict, each_pt_results: List[RowDict]) -> PitchFxMetricsCollection:
        metrics_collection = _get_pitchfx_metrics_for_all_pitch_types(all_pt_results)
        metrics_collection["pitch_type_metrics"] = _get_pitchfx_metrics_for_each_pitch_type(each_pt_results)
        metrics_collection["pitch_type"] = _get_pfx_collection_pitch_type(metrics_collection["pitch_type_metrics"])
        return from_dict(data_class=cls, data=metrics_collection)


def _get_pitchfx_metrics_for_all_pitch_types(results: RowDict) -> PitchFxMetricsDict:
    metrics_collection = PitchFxMetrics._get_pitchfx_metrics(results)
    metrics_collection["percent"] = 1.0
    return metrics_collection


def _get_pitchfx_metrics_for_each_pitch_type(results: List[RowDict]) -> PitchTypeMetricsDict:
    all_pitch_types = [PitchFxMetrics._get_pitchfx_metrics(pitch_type_metrics) for pitch_type_metrics in results]
    total_pitches = sum(metrics["total_pitches"] for metrics in all_pitch_types)
    for pfx_metrics in all_pitch_types:
        pfx_metrics["percent"] = round(pfx_metrics["total_pitches"] / total_pitches, 3)
    all_pitch_types = list(filter(lambda x: x["percent"] >= 0.01, all_pitch_types))
    return {m["pitch_type"]: m for m in sorted(all_pitch_types, key=lambda x: x["percent"], reverse=True)}


def _get_pfx_collection_pitch_type(pitch_type_metrics: PitchTypeMetricsDict) -> PitchType:
    all_pitch_types_int = sum(int(pitch_type) for pitch_type in list(pitch_type_metrics.keys()))
    return PitchType(all_pitch_types_int)



# flake8: noqa
"""Installation script for vigorish CLI application."""
from pathlib import Path

from setuptools import find_packages, setup

DESCRIPTION = "Hybrid Python/Node.js web scraper for Major League Baseball (MLB) data."
APP_ROOT = Path(__file__).parent
README = (APP_ROOT / "README.md").read_text().strip()
AUTHOR = "Aaron Luna"
AUTHOR_EMAIL = "contact@aaronluna.dev"
PROJECT_URLS = {
    "Bug Tracker": "https://github.com/a-luna/vigorish/issues",
    "Source Code": "https://github.com/a-luna/vigorish",
    "Documentation": "https://aaronluna.dev/projects/vigorish",
}
CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "Environment :: Console",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: MIT License",
    "Natural Language :: English",
    "Operating System :: MacOS :: MacOS X",
    "Operating System :: POSIX :: Linux",
    "Operating System :: Unix",
    "Programming Language :: JavaScript",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3 :: Only",
]
INSTALL_REQUIRES = [
    "aenum",
    "boto3",
    "bullet",
    "click==7.1.2",
    "dacite",
    "dataclass-csv",
    "enlighten",
    "events",
    "halo",
    "lxml",
    "naked",
    "py-getch",
    "pyfiglet",
    "python-dateutil",
    "python-dotenv",
    "rapidfuzz",
    "requests",
    "scipy",
    "sqlalchemy==1.3.23",
    "sqlalchemy-utils",
    "tabulate",
    "tqdm",
    "urllib3",
    "w3lib",
]

exec(open(str(APP_ROOT / "src/vigorish/version.py")).read())
setup(
    name="vigorish",
    description=DESCRIPTION,
    long_description=README,
    long_description_content_type="text/markdown",
    version=__version__,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    maintainer=AUTHOR,
    maintainer_email=AUTHOR_EMAIL,
    license="MIT",
    url=PROJECT_URLS["Source Code"],
    project_urls=PROJECT_URLS,
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    python_requires=">=3.6",
    classifiers=CLASSIFIERS,
    install_requires=INSTALL_REQUIRES,
    entry_points="""
        [console_scripts]
        vig=vigorish.cli.vig:cli
    """,
)



"""Functions that enable reading/writing the config file."""
from abc import ABC
from typing import Iterable, Mapping, Union

from vigorish.config.types.batch_job_settings import BatchJobSettings
from vigorish.config.types.batch_scrape_delay import BatchScrapeDelay
from vigorish.config.types.folder_path_setting import LocalFolderPathSetting, S3FolderPathSetting
from vigorish.config.types.url_scrape_delay import UrlScrapeDelay
from vigorish.enums import (
    CombinedDataStorageOption,
    ConfigType,
    DataSet,
    HtmlStorageOption,
    JsonStorageOption,
    ScrapeCondition,
    ScrapeTaskOption,
    StatusReport,
)
from vigorish.util.list_helpers import report_dict


def same_value_for_all_data_sets_is_required(setting_name: str) -> bool:
    return setting_name in [
        "STATUS_REPORT",
        "S3_BUCKET",
        "SCRAPE_TASK_OPTION",
        "SCRAPED_DATA_COMBINE_CONDITION",
        "COMBINED_DATA_STORAGE",
        "COMBINED_DATA_LOCAL_FOLDER_PATH",
        "COMBINED_DATA_S3_FOLDER_PATH",
        "DB_BACKUP_FOLDER_PATH",
    ]


EnumSetting = Union[
    CombinedDataStorageOption,
    HtmlStorageOption,
    JsonStorageOption,
    ScrapeCondition,
    ScrapeTaskOption,
]
NumericSetting = Union[BatchJobSettings, BatchScrapeDelay, UrlScrapeDelay]
PathSetting = Union[LocalFolderPathSetting, S3FolderPathSetting]
ConfigValue = Union[None, bool, int, str, EnumSetting]
ConfigDict = Mapping[str, Union[None, bool, int, str]]
ConfigSettingValue = Union[EnumSetting, NumericSetting, PathSetting, str]


class ConfigSetting(ABC):
    data_type: ConfigType
    setting_name: str
    config_dict: ConfigDict

    def __init__(self, setting_name: str, config_dict: ConfigDict):
        self.setting_name = setting_name
        self.config_dict = config_dict
        self.data_type = ConfigType.NONE

    @property
    def setting_name_title(self) -> str:
        return " ".join(self.setting_name.split("_")).title()

    @property
    def description(self) -> str:
        return self.config_dict.get("DESCRIPTION", "")

    @property
    def possible_values(self) -> Iterable[EnumSetting]:
        return []

    @property
    def is_same_for_all_data_sets(self) -> bool:
        return (
            True
            if same_value_for_all_data_sets_is_required(self.setting_name)
            else self.config_dict.get("SAME_SETTING_FOR_ALL_DATA_SETS")
        )

    @property
    def current_settings_report(self) -> str:
        if self.is_same_for_all_data_sets:
            settings_dict = {"ALL_DATA_SETS": self.current_setting(DataSet.ALL)}
        else:
            settings_dict = {ds.name: self.current_setting(ds) for ds in DataSet}
        return report_dict(settings_dict, title="", title_prefix="", title_suffix="")

    def current_setting(self, data_set: DataSet) -> ConfigValue:
        return (
            self.config_dict.get(DataSet.ALL.name)
            if self.is_same_for_all_data_sets
            else self.config_dict.get(data_set.name)
        )


class StringConfigSetting(ConfigSetting):
    def __init__(self, setting_name: str, config_dict: ConfigDict):
        super().__init__(setting_name, config_dict)
        self.data_type = ConfigType.STRING

    def __repr__(self) -> str:
        return f"<StringConfigSetting setting={self.setting_name}, value={self.current_settings_report}>"


class PathConfigSetting(ConfigSetting):
    def __init__(self, setting_name: str, config_dict: ConfigDict):
        super().__init__(setting_name, config_dict)
        self.data_type = ConfigType.PATH

    def __repr__(self) -> str:
        return f"<PathConfigSetting setting={self.setting_name}, value={self.current_settings_report}>"

    def current_setting(self, data_set: DataSet) -> PathSetting:
        current_setting = super().current_setting(data_set)
        class_name = self.config_dict.get("CLASS_NAME")
        if class_name == "LocalFolderPathSetting":
            return LocalFolderPathSetting(path=current_setting, data_set=data_set)
        if class_name == "S3FolderPathSetting":
            return S3FolderPathSetting(path=current_setting, data_set=data_set)


class EnumConfigSetting(ConfigSetting):
    def __init__(self, setting_name: str, config_dict: ConfigDict):
        super().__init__(setting_name, config_dict)
        self.data_type = ConfigType.ENUM

    def __repr__(self) -> str:
        return f"<EnumConfigSetting setting={self.setting_name}, value={self.current_settings_report}>"

    @property
    def enum_name(self) -> str:
        return self.config_dict.get("ENUM_NAME")

    @property
    def possible_values(self) -> Iterable[EnumSetting]:
        if self.enum_name == "ScrapeCondition":
            return list(ScrapeCondition)
        if self.enum_name == "ScrapeTaskOption":
            return list(ScrapeTaskOption)
        if self.enum_name == "HtmlStorageOption":
            return list(HtmlStorageOption)
        if self.enum_name == "JsonStorageOption":
            return list(JsonStorageOption)
        if self.enum_name == "CombinedDataStorageOption":
            return list(CombinedDataStorageOption)
        if self.enum_name == "StatusReport":
            return list(StatusReport)
        return []

    def current_setting(self, data_set: DataSet) -> EnumSetting:
        current_setting = super().current_setting(data_set).upper()
        if self.enum_name == "ScrapeCondition":
            return ScrapeCondition[current_setting]
        if self.enum_name == "ScrapeTaskOption":
            return ScrapeTaskOption[current_setting]
        if self.enum_name == "HtmlStorageOption":
            return HtmlStorageOption[current_setting]
        if self.enum_name == "JsonStorageOption":
            return JsonStorageOption[current_setting]
        if self.enum_name == "CombinedDataStorageOption":
            return CombinedDataStorageOption[current_setting]
        if self.enum_name == "StatusReport":
            return StatusReport[current_setting]
        return None


class NumericConfigSetting(ConfigSetting):
    def __init__(self, setting_name: str, config_dict: ConfigDict):
        super().__init__(setting_name, config_dict)
        self.data_type = ConfigType.NUMERIC

    def __repr__(self) -> str:
        return f"<NumericConfigSetting setting={self.setting_name}, value={self.current_settings_report}>"

    def current_setting(self, data_set: DataSet) -> NumericSetting:
        current_setting = super().current_setting(data_set)
        class_name = self.config_dict.get("CLASS_NAME")
        if class_name == "UrlScrapeDelay":
            return UrlScrapeDelay.from_config(current_setting)
        if class_name == "BatchJobSettings":
            return BatchJobSettings.from_config(current_setting)
        if class_name == "BatchScrapeDelay":
            return BatchScrapeDelay.from_config(current_setting)
        return None



import vigorish.database as db
from vigorish.enums import DataSet
from vigorish.util.dt_format_strings import DATE_ONLY, DATE_ONLY_TABLE_ID
from vigorish.util.result import Result


def update_bbref_games_for_date_single_date(db_session, season, games_for_date):
    result = update_status_bbref_games_for_date(db_session, games_for_date)
    if result.failure:
        return result
    result = create_game_status_records(db_session, season, games_for_date)
    if result.failure:
        return result
    db_session.commit()
    return Result.Ok()


def update_bbref_games_for_date_list(scraped_data, db_session, scraped_bbref_dates, apply_patch_list=True):
    season = None
    for game_date in sorted(scraped_bbref_dates):
        if not season:
            season = db.Season.find_by_year(db_session, game_date.year)
        games_for_date = scraped_data.get_bbref_games_for_date(game_date, apply_patch_list)
        if not games_for_date:
            date_str = game_date.strftime(DATE_ONLY_TABLE_ID)
            error = f"Failed to retrieve {DataSet.BBREF_GAMES_FOR_DATE} (URL ID: {date_str})"
            return Result.Fail(error)
        result = update_status_bbref_games_for_date(db_session, games_for_date)
        if result.failure:
            return result
        result = create_game_status_records(db_session, season, games_for_date)
        if result.failure:
            return result
        db_session.commit()
    return Result.Ok()


def update_status_bbref_games_for_date(db_session, games_for_date):
    try:
        date_id = games_for_date.game_date.strftime(DATE_ONLY_TABLE_ID)
        date_status = db_session.query(db.DateScrapeStatus).get(date_id)
        if not date_status:
            date_str = games_for_date.game_date.strftime(DATE_ONLY)
            error = f"scrape_status_date does not contain an entry for date: {date_str}"
            return Result.Fail(error)
        date_status.scraped_daily_dash_bbref = 1
        date_status.game_count_bbref = games_for_date.game_count
        return Result.Ok()
    except Exception as e:
        return Result.Fail(f"Error: {repr(e)}")


def create_game_status_records(db_session, season, games_for_date):
    new_bbref_game_ids = games_for_date.all_bbref_game_ids
    game_status_bbref_ids = db.GameScrapeStatus.get_all_bbref_game_ids(db_session, season.id)
    missing_bbref_game_ids = set(new_bbref_game_ids).difference(set(game_status_bbref_ids))
    if not missing_bbref_game_ids:
        return Result.Ok()
    for game_info in filter(lambda x: x.bbref_game_id in missing_bbref_game_ids, games_for_date.games):
        try:
            game_date = games_for_date.game_date
            date_status = db.DateScrapeStatus.find_by_date(db_session, game_date)
            game_status = db.GameScrapeStatus()
            game_status.game_date = game_date
            game_status.bbref_game_id = game_info.bbref_game_id
            game_status.scrape_status_date_id = date_status.id
            game_status.season_id = season.id
            db_session.add(game_status)
        except Exception as e:
            return Result.Fail(f"Error: {repr(e)}")
    return Result.Ok()



"""Functions which operate on or produce list objects."""
import itertools
from collections import defaultdict, OrderedDict
from datetime import datetime
from random import randint
from typing import Dict

from vigorish.util.dt_format_strings import DT_AWARE


def display_dict(dict, title="", title_prefix="[", title_suffix="]", extra_dots=2):
    """Pretty print all dict keys and values, title is optional."""
    report = report_dict(
        dict=dict,
        title=title,
        title_prefix=title_prefix,
        title_suffix=title_suffix,
        extra_dots=extra_dots,
    )
    print(report)


def report_dict(dict, title="", title_prefix="### ", title_suffix=" ###", extra_dots=2):
    """Pretty print all dict keys and values, title is optional."""
    report = "\n"
    max_length = 0
    for k, v in dict.items():
        if not v and type(v) is not int and type(v) is not bool:
            continue
        if len(k) > max_length:
            max_length = len(k)
    max_length += extra_dots

    if title:
        report += f"{title_prefix}{title}{title_suffix}"
    for k, v in dict.items():
        if not v and type(v) is not int and type(v) is not bool:
            continue
        if type(v) is datetime:
            v = v.strftime(DT_AWARE)
        c = max_length - len(k)
        report += f"\n{k}{'.' * c}: {v}"
    return report.strip()


def dict_to_param_list(input_dict: Dict):
    params = []
    for name, value in input_dict.items():
        if isinstance(value, bool) and value:
            params.append(f"--{name}")
        else:
            params.append(f"--{name}={value:}")
    return " ".join(params)


def as_dict_list(db_objects):
    """Convert a list of model objects to a list of dicts."""
    return [obj.as_dict() for obj in db_objects if hasmethod(obj, "as_dict")]


def hasmethod(obj, method_name):
    """Return True if obj.method_name exists and is callable. Otherwise, return False."""
    obj_method = getattr(obj, method_name, None)
    return callable(obj_method) if obj_method else False


def flatten_list2d(list2d):
    """Produce a normal list by flattenning a 2-dimensional list."""
    return list(itertools.chain(*list2d))


def group_and_sort_list(unsorted, group_attr, sort_attr, sort_groups_desc=False, sort_all_desc=False):
    list_sorted = sorted(unsorted, key=lambda x: getattr(x, sort_attr), reverse=sort_all_desc)
    list_grouped = defaultdict(list)
    for item in list_sorted:
        list_grouped[getattr(item, group_attr)].append(item)
    grouped_sorted = OrderedDict()
    for group in sorted(list_grouped.keys(), reverse=sort_groups_desc):
        grouped_sorted[group] = list_grouped[group]
    return grouped_sorted


def group_and_sort_dict_list(unsorted, group_key, sort_key, sort_groups_desc=False, sort_all_desc=False):
    list_sorted = sorted(unsorted, key=lambda x: x[sort_key], reverse=sort_all_desc)
    list_grouped = defaultdict(list)
    for item in list_sorted:
        list_grouped[item[group_key]].append(item)
    grouped_sorted = OrderedDict()
    for group in sorted(list_grouped.keys(), reverse=sort_groups_desc):
        grouped_sorted[group] = list_grouped[group]
    return grouped_sorted


def compare_lists(list1, list2):
    check1 = not list(set(list1) - set(list2))
    check2 = not list(set(list2) - set(list1))
    return check1 and check2


def make_chunked_list(input_list, chunk_size):
    chunked_list = []
    (total_chunks, last_chunk_size) = divmod(len(input_list), chunk_size)
    if last_chunk_size:
        total_chunks += 1
    for i in range(total_chunks):
        start = i * chunk_size
        end = start + chunk_size
        if last_chunk_size and i == total_chunks - 1:
            end = len(input_list)
        chunked_list.append(input_list[start:end])
    return chunked_list


def make_irregular_chunked_list(input_list, min_chunk_size, max_chunk_size):
    chunked_list = []
    remaining = len(input_list)
    end = 0
    while remaining:
        if remaining > max_chunk_size:
            chunk_size = randint(min_chunk_size, max_chunk_size)
        else:
            chunk_size = remaining
        start = end
        end += chunk_size
        chunked_list.append(input_list[start:end])
        remaining -= chunk_size
    return chunked_list



from sqlalchemy import and_, func, join, select
from sqlalchemy_utils import create_view

import vigorish.database as db
from vigorish.enums import SeasonType
from vigorish.util.list_helpers import flatten_list2d


class Season_Game_PitchApp_View(db.Base):
    __table__ = create_view(
        name="season_game_pitch_app_status",
        selectable=select(
            [
                db.Season.id.label("id"),
                db.Season.year.label("year"),
                db.Season.season_type.label("season_type"),
                db.PitchAppScrapeStatus.scrape_status_date_id.label("date_id"),
                db.PitchAppScrapeStatus.scrape_status_game_id.label("game_id"),
                db.PitchAppScrapeStatus.bbref_game_id.label("bbref_game_id"),
                func.count(db.PitchAppScrapeStatus.id).label("total_pitchfx"),
                func.sum(db.PitchAppScrapeStatus.scraped_pitchfx).label("total_pitchfx_scraped"),
                func.sum(db.PitchAppScrapeStatus.no_pitchfx_data).label("total_no_pitchfx_data"),
                func.sum(db.PitchAppScrapeStatus.combined_pitchfx_bbref_data).label(
                    "total_combined_pitchfx_bbref_data"
                ),
                func.sum(db.PitchAppScrapeStatus.batters_faced_bbref).label("total_batters_faced_bbref"),
                func.sum(db.PitchAppScrapeStatus.batters_faced_pitchfx).label("total_batters_faced_pitchfx"),
                func.sum(db.PitchAppScrapeStatus.pitch_count_pitch_log).label("total_pitch_count_pitch_log"),
                func.sum(db.PitchAppScrapeStatus.pitch_count_bbref).label("total_pitch_count_bbref"),
                func.sum(db.PitchAppScrapeStatus.pitch_count_pitchfx).label("total_pitch_count_pitchfx"),
                func.sum(db.PitchAppScrapeStatus.pitch_count_pitchfx_audited).label(
                    "total_pitch_count_pitchfx_audited"
                ),
                func.sum(db.PitchAppScrapeStatus.total_at_bats_pitchfx_complete).label(
                    "total_at_bats_pitchfx_complete"
                ),
                func.sum(db.PitchAppScrapeStatus.patched_pitchfx_count).label("total_patched_pitchfx_count"),
                func.sum(db.PitchAppScrapeStatus.total_at_bats_patched_pitchfx).label("total_at_bats_patched_pitchfx"),
                func.sum(db.PitchAppScrapeStatus.missing_pitchfx_count).label("total_missing_pitchfx_count"),
                func.sum(db.PitchAppScrapeStatus.total_at_bats_missing_pitchfx).label("total_at_bats_missing_pitchfx"),
                func.sum(db.PitchAppScrapeStatus.removed_pitchfx_count).label("total_removed_pitchfx_count"),
                func.sum(db.PitchAppScrapeStatus.total_at_bats_removed_pitchfx).label("total_at_bats_removed_pitchfx"),
                func.sum(db.PitchAppScrapeStatus.pitchfx_error).label("total_pitchfx_error"),
                func.sum(db.PitchAppScrapeStatus.total_at_bats_pitchfx_error).label("total_at_bats_pitchfx_error"),
                func.sum(db.PitchAppScrapeStatus.invalid_pitchfx).label("total_invalid_pitchfx"),
                func.sum(db.PitchAppScrapeStatus.invalid_pitchfx_count).label("total_invalid_pitchfx_count"),
                func.sum(db.PitchAppScrapeStatus.total_at_bats_invalid_pitchfx).label("total_at_bats_invalid_pitchfx"),
            ]
        )
        .select_from(
            join(
                db.Season,
                db.PitchAppScrapeStatus,
                db.Season.id == db.PitchAppScrapeStatus.season_id,
            )
        )
        .group_by(db.PitchAppScrapeStatus.scrape_status_game_id),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_all_bbref_game_ids_eligible_for_audit(cls, db_engine, year, season_type=SeasonType.REGULAR_SEASON):
        s = select([cls.bbref_game_id]).where(
            and_(
                cls.year == year,
                cls.season_type == season_type,
                cls.total_pitchfx == cls.total_pitchfx_scraped,
                cls.total_pitchfx != cls.total_combined_pitchfx_bbref_data,
            )
        )
        results = db_engine.execute(s).fetchall()
        return flatten_list2d([d.values() for d in [dict(row) for row in results]])

    @classmethod
    def get_all_bbref_game_ids_all_pitchfx_logs_are_valid(cls, db_engine, year, season_type=SeasonType.REGULAR_SEASON):
        s = select([cls.bbref_game_id]).where(
            and_(
                cls.year == year,
                cls.season_type == season_type,
                cls.total_pitchfx == cls.total_combined_pitchfx_bbref_data,
                cls.total_pitchfx_error == 0,
                cls.total_invalid_pitchfx == 0,
            )
        )
        results = db_engine.execute(s).fetchall()
        return flatten_list2d([d.values() for d in [dict(row) for row in results]])

    @classmethod
    def get_all_bbref_game_ids_pitchfx_error(cls, db_engine, year, season_type=SeasonType.REGULAR_SEASON):
        s = select([cls.bbref_game_id]).where(
            and_(
                cls.year == year,
                cls.season_type == season_type,
                cls.total_pitchfx_error > 0,
            )
        )
        results = db_engine.execute(s).fetchall()
        return flatten_list2d([d.values() for d in [dict(row) for row in results]])

    @classmethod
    def get_all_bbref_game_ids_invalid_pitchfx(cls, db_engine, year, season_type=SeasonType.REGULAR_SEASON):
        s = select([cls.bbref_game_id]).where(
            and_(
                cls.year == year,
                cls.season_type == season_type,
                cls.total_invalid_pitchfx > 0,
            )
        )
        results = db_engine.execute(s).fetchall()
        return flatten_list2d([d.values() for d in [dict(row) for row in results]])

    @classmethod
    def get_all_bbref_game_ids_combined_no_missing_pfx(cls, db_engine, season_type=SeasonType.REGULAR_SEASON):
        s = select([cls.bbref_game_id]).where(
            and_(
                cls.season_type == season_type,
                cls.total_pitchfx == cls.total_combined_pitchfx_bbref_data,
                cls.total_missing_pitchfx_count == 0,
                cls.total_pitchfx_error == 0,
                cls.total_invalid_pitchfx == 0,
            )
        )
        results = db_engine.execute(s).fetchall()
        return flatten_list2d([d.values() for d in [dict(row) for row in results]])



"""Association table linking Players to Teams."""
from sqlalchemy import Column, ForeignKey, Integer, String

import vigorish.database as db


class Assoc_Player_Team(db.Base):
    __tablename__ = "player_team"

    db_player_id = Column(Integer, ForeignKey("player.id"), primary_key=True)
    db_team_id = Column(Integer, ForeignKey("team.id"), primary_key=True)
    mlb_id = Column(Integer)
    bbref_id = Column(String)
    team_id = Column(String)
    year = Column(Integer)
    stint_number = Column(Integer)
    season_id = Column(Integer, ForeignKey("season.id"))



"""Menu item that allows the user to initialize/reset the database."""
import subprocess
import time
from collections import defaultdict
from datetime import datetime
from pathlib import Path

from getch import pause
from halo import Halo

import vigorish.database as db
from vigorish.cli.components.prompts import yes_no_prompt
from vigorish.cli.components.util import (
    get_random_cli_color,
    get_random_dots_spinner,
    print_error,
    print_message,
    shutdown_cli_immediately,
)
from vigorish.cli.menu_item import MenuItem
from vigorish.cli.menu_items.admin_tasks.update_player_id_map import UpdatePlayerIdMap
from vigorish.constants import EMOJIS
from vigorish.enums import DataSet
from vigorish.tasks import ImportScrapedDataTask
from vigorish.util.result import Result

SETUP_HEADING = (
    "Before you can begin scraping data, you must initialize the database with initial player, team and season data."
)
SETUP_MESSAGE = "\nSelect YES to initialize the database\nSelect NO to return to the previous menu"
RESET_MESSAGE = "\nWould you like to reset the database with initial player, team and season data?"
WARNING = (
    "WARNING! All existing data will be deleted if you choose to reset the database. This action cannot be undone."
)
IMPORT_SCRAPED_DATA_MESSAGE = (
    "Would you like to update the database using files from previous vigorish installations? "
    "Files from any MLB season and any data set found in the location specified by the "
    "JSON_LOCAL_FOLDER_PATH config setting will be used for this step.\n"
)
IMPORT_SCRAPED_DATA_PROMPT = (
    "\nSelect YES to import all data in the location above\n" "Select NO to return to the previous menu"
)
DB_INITIALIZED = "Database has been successfully initialized.\n"


class SetupDatabase(MenuItem):
    def __init__(self, app):
        super().__init__(app)
        self.update_id_map_task = UpdatePlayerIdMap(self.app)
        self.import_data_task = ImportScrapedDataTask(self.app)
        self.db_initialized = self.app.db_setup_complete
        self.spinners = {}
        self.menu_item_text = "Reset Database" if self.db_initialized else "Setup Database"
        self.menu_item_emoji = EMOJIS["BOMB"] if self.db_initialized else EMOJIS["DIZZY"]
        self.menu_heading = self._menu_item_text

    def launch(self):
        restart_required = self.db_initialized
        if not self.prompt_user_run_task():
            return Result.Ok(self.exit_menu)
        return (
            self.update_player_id_map()
            .on_success(self.update_database_connection)
            .on_success(self.create_and_populate_database_tables)
            .on_success(self.import_scraped_data, restart_required)
        )

    def prompt_user_run_task(self):
        if self.db_initialized:
            message = WARNING
            message_color = "bright_red"
            prompt = RESET_MESSAGE
        else:
            message = SETUP_HEADING
            message_color = None
            prompt = SETUP_MESSAGE
        self.update_menu_heading("Run Task?")
        print_message(message, fg=message_color)
        return yes_no_prompt(prompt, wrap=False)

    def update_player_id_map(self):
        if not self.db_initialized:
            return Result.Ok()
        result = db.Season.is_date_in_season(self.db_session, datetime.now())
        if result.failure:
            return Result.Ok()
        subprocess.run(["clear"])
        return self.update_id_map_task.launch()

    def update_database_connection(self):
        if self.db_initialized:
            self.app.reset_database_connection()
        return Result.Ok()

    def create_and_populate_database_tables(self):
        self.update_menu_heading("In Progress")
        result = self.app.initialize_database()
        if result.success:
            self.update_menu_heading("Complete!", heading_color="bright_green")
            print_message(DB_INITIALIZED, fg="bright_green", bold=True)
            pause(message="Press any key to continue...")
        return result

    def import_scraped_data(self, restart_required):
        if not self.import_scraped_data_prompt():
            return self.setup_complete(restart_required)
        subprocess.run(["clear"])
        self.subscribe_to_events()
        result = self.import_data_task.execute(overwrite_existing=True)
        if result.error:
            print_message(result.error, fg="bright_red")
            pause(message="Press any key to continue...")
        else:
            time.sleep(2)
        self.unsubscribe_from_events()
        return self.setup_complete(restart_required)

    def import_scraped_data_prompt(self):
        example_folder = self.app.get_current_setting("JSON_LOCAL_FOLDER_PATH", DataSet.BBREF_BOXSCORES, year=2019)
        root_folder = Path(example_folder).parent.parent
        current_setting = f"JSON_LOCAL_FOLDER_PATH: {root_folder}"
        self.update_menu_heading("Import Local JSON Folder?")
        print_message(IMPORT_SCRAPED_DATA_MESSAGE)
        print_message(current_setting, wrap=False)
        return yes_no_prompt(IMPORT_SCRAPED_DATA_PROMPT, wrap=False)

    def setup_complete(self, restart_required):
        if restart_required:
            shutdown_cli_immediately()
        return Result.Ok(self.exit_menu)

    def error_occurred(self, error_message, data_set, year):
        self.update_menu_heading("Error!", heading_color="bright_red")
        self.spinner.fail(f"Error occurred while updating {data_set} for MLB {year}")
        print_error(f"\nError: {error_message}")

    def search_local_files_start(self):
        self.update_menu_heading("In Progress")
        self.spinners["default"] = Halo(spinner=get_random_dots_spinner(), color=get_random_cli_color())
        self.spinners["default"].text = "Searching local folder for scraped data..."
        self.spinners["default"].start()

    def import_scraped_data_start(self):
        self.spinners["default"].stop()
        self.update_menu_heading("In Progress...")

    def import_scraped_data_complete(self):
        self.update_menu_heading("Complete!", heading_color="bright_green")
        success = "Successfully imported all scraped data from local files"
        print_message(success, fg="bright_yellow", bold=True)
        pause(message="\nPress any key to continue...")

    def import_scraped_data_for_year_start(self, year):
        self.update_menu_heading("In Progress...")
        self.spinners[year] = defaultdict(lambda: Halo())

    def import_scraped_data_set_start(self, data_set, year):
        spinner = self.spinners[year][data_set]
        spinner.spinner = get_random_dots_spinner()
        spinner.color = get_random_cli_color()
        spinner.text = f"Updating {data_set} for MLB {year}..."
        spinner.start()

    def import_scraped_data_set_complete(self, data_set, year):
        self.spinners[year][data_set].succeed(f"Successfully updated {data_set} for MLB {year}!")

    def subscribe_to_events(self):
        self.import_data_task.events.error_occurred += self.error_occurred
        self.import_data_task.events.search_local_files_start += self.search_local_files_start
        self.import_data_task.events.import_scraped_data_start += self.import_scraped_data_start
        self.import_data_task.events.import_scraped_data_complete += self.import_scraped_data_complete
        self.import_data_task.events.import_scraped_data_for_year_start += self.import_scraped_data_for_year_start
        self.import_data_task.events.import_scraped_data_set_start += self.import_scraped_data_set_start
        self.import_data_task.events.import_scraped_data_set_complete += self.import_scraped_data_set_complete

    def unsubscribe_from_events(self):
        self.import_data_task.events.error_occurred -= self.error_occurred
        self.import_data_task.events.search_local_files_start -= self.search_local_files_start
        self.import_data_task.events.import_scraped_data_start -= self.import_scraped_data_start
        self.import_data_task.events.import_scraped_data_complete -= self.import_scraped_data_complete
        self.import_data_task.events.import_scraped_data_for_year_start -= self.import_scraped_data_for_year_start
        self.import_data_task.events.import_scraped_data_set_start -= self.import_scraped_data_set_start
        self.import_data_task.events.import_scraped_data_set_complete -= self.import_scraped_data_set_complete



from sqlalchemy import and_, func, join, select
from sqlalchemy_utils.view import create_view

import vigorish.database as db
from vigorish.data.metrics.bat_stats import BatStatsMetrics
from vigorish.views.bat_stats_col_expressions import (
    avg,
    bb_rate,
    contact_rate,
    iso,
    k_rate,
    obp,
    ops,
    slg,
)


class BatStats_All_View(db.Base):
    __table__ = create_view(
        name="batstats_all",
        selectable=select(
            [
                db.BatStats.player_id.label("id"),
                db.BatStats.player_id_mlb.label("mlb_id"),
                db.BatStats.player_id_bbref.label("bbref_id"),
                func.count(db.BatStats.id).label("total_games"),
                avg,
                obp,
                slg,
                ops,
                iso,
                bb_rate,
                k_rate,
                contact_rate,
                func.sum(db.BatStats.plate_appearances).label("plate_appearances"),
                func.sum(db.BatStats.at_bats).label("at_bats"),
                func.sum(db.BatStats.hits).label("hits"),
                func.sum(db.BatStats.runs_scored).label("runs_scored"),
                func.sum(db.BatStats.rbis).label("rbis"),
                func.sum(db.BatStats.bases_on_balls).label("bases_on_balls"),
                func.sum(db.BatStats.strikeouts).label("strikeouts"),
                func.sum(db.BatStats.doubles).label("doubles"),
                func.sum(db.BatStats.triples).label("triples"),
                func.sum(db.BatStats.homeruns).label("homeruns"),
                func.sum(db.BatStats.stolen_bases).label("stolen_bases"),
                func.sum(db.BatStats.caught_stealing).label("caught_stealing"),
                func.sum(db.BatStats.hit_by_pitch).label("hit_by_pitch"),
                func.sum(db.BatStats.intentional_bb).label("intentional_bb"),
                func.sum(db.BatStats.gdp).label("gdp"),
                func.sum(db.BatStats.sac_fly).label("sac_fly"),
                func.sum(db.BatStats.sac_hit).label("sac_hit"),
                func.sum(db.BatStats.total_pitches).label("total_pitches"),
                func.sum(db.BatStats.total_strikes).label("total_strikes"),
                func.sum(db.BatStats.wpa_bat).label("wpa_bat"),
                func.sum(db.BatStats.wpa_bat_pos).label("wpa_bat_pos"),
                func.sum(db.BatStats.wpa_bat_neg).label("wpa_bat_neg"),
                func.sum(db.BatStats.re24_bat).label("re24_bat"),
            ]
        )
        .select_from(db.BatStats)
        .group_by(db.BatStats.player_id)
        .order_by(db.BatStats.player_id_mlb),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_bat_stats_for_career_for_player(cls, db_engine, mlb_id):
        s = select([cls]).where(cls.mlb_id == mlb_id)
        results = db_engine.execute(s).fetchall()
        return BatStatsMetrics.from_query_results(results)[0] if results else None


class BatStats_By_Year_View(db.Base):
    __table__ = create_view(
        name="batstats_by_year",
        selectable=select(
            [
                db.BatStats.player_id.label("id"),
                db.BatStats.player_id_mlb.label("mlb_id"),
                db.BatStats.player_id_bbref.label("bbref_id"),
                db.BatStats.season_id.label("season_id"),
                db.Season.year.label("year"),
                func.count(db.BatStats.id).label("total_games"),
                avg,
                obp,
                slg,
                ops,
                iso,
                bb_rate,
                k_rate,
                contact_rate,
                func.sum(db.BatStats.plate_appearances).label("plate_appearances"),
                func.sum(db.BatStats.at_bats).label("at_bats"),
                func.sum(db.BatStats.hits).label("hits"),
                func.sum(db.BatStats.runs_scored).label("runs_scored"),
                func.sum(db.BatStats.rbis).label("rbis"),
                func.sum(db.BatStats.bases_on_balls).label("bases_on_balls"),
                func.sum(db.BatStats.strikeouts).label("strikeouts"),
                func.sum(db.BatStats.doubles).label("doubles"),
                func.sum(db.BatStats.triples).label("triples"),
                func.sum(db.BatStats.homeruns).label("homeruns"),
                func.sum(db.BatStats.stolen_bases).label("stolen_bases"),
                func.sum(db.BatStats.caught_stealing).label("caught_stealing"),
                func.sum(db.BatStats.hit_by_pitch).label("hit_by_pitch"),
                func.sum(db.BatStats.intentional_bb).label("intentional_bb"),
                func.sum(db.BatStats.gdp).label("gdp"),
                func.sum(db.BatStats.sac_fly).label("sac_fly"),
                func.sum(db.BatStats.sac_hit).label("sac_hit"),
                func.sum(db.BatStats.total_pitches).label("total_pitches"),
                func.sum(db.BatStats.total_strikes).label("total_strikes"),
                func.sum(db.BatStats.wpa_bat).label("wpa_bat"),
                func.sum(db.BatStats.wpa_bat_pos).label("wpa_bat_pos"),
                func.sum(db.BatStats.wpa_bat_neg).label("wpa_bat_neg"),
                func.sum(db.BatStats.re24_bat).label("re24_bat"),
            ]
        )
        .select_from(join(db.BatStats, db.Season, db.BatStats.season_id == db.Season.id))
        .group_by(db.BatStats.season_id)
        .group_by(db.BatStats.player_id)
        .order_by(db.BatStats.player_id_mlb),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_bat_stats_by_year_for_player(cls, db_engine, mlb_id):
        s = select([cls]).where(cls.mlb_id == mlb_id)
        results = db_engine.execute(s).fetchall()
        return BatStatsMetrics.from_query_results(results) if results else []


class BatStats_By_Team_View(db.Base):
    __table__ = create_view(
        name="batstats_by_team",
        selectable=select(
            [
                db.BatStats.player_id.label("id"),
                db.BatStats.player_id_mlb.label("mlb_id"),
                db.BatStats.player_id_bbref.label("bbref_id"),
                db.BatStats.player_team_id_bbref.label("player_team_id_bbref"),
                func.count(db.BatStats.id).label("total_games"),
                avg,
                obp,
                slg,
                ops,
                iso,
                bb_rate,
                k_rate,
                contact_rate,
                func.sum(db.BatStats.plate_appearances).label("plate_appearances"),
                func.sum(db.BatStats.at_bats).label("at_bats"),
                func.sum(db.BatStats.hits).label("hits"),
                func.sum(db.BatStats.runs_scored).label("runs_scored"),
                func.sum(db.BatStats.rbis).label("rbis"),
                func.sum(db.BatStats.bases_on_balls).label("bases_on_balls"),
                func.sum(db.BatStats.strikeouts).label("strikeouts"),
                func.sum(db.BatStats.doubles).label("doubles"),
                func.sum(db.BatStats.triples).label("triples"),
                func.sum(db.BatStats.homeruns).label("homeruns"),
                func.sum(db.BatStats.stolen_bases).label("stolen_bases"),
                func.sum(db.BatStats.caught_stealing).label("caught_stealing"),
                func.sum(db.BatStats.hit_by_pitch).label("hit_by_pitch"),
                func.sum(db.BatStats.intentional_bb).label("intentional_bb"),
                func.sum(db.BatStats.gdp).label("gdp"),
                func.sum(db.BatStats.sac_fly).label("sac_fly"),
                func.sum(db.BatStats.sac_hit).label("sac_hit"),
                func.sum(db.BatStats.total_pitches).label("total_pitches"),
                func.sum(db.BatStats.total_strikes).label("total_strikes"),
                func.sum(db.BatStats.wpa_bat).label("wpa_bat"),
                func.sum(db.BatStats.wpa_bat_pos).label("wpa_bat_pos"),
                func.sum(db.BatStats.wpa_bat_neg).label("wpa_bat_neg"),
                func.sum(db.BatStats.re24_bat).label("re24_bat"),
            ]
        )
        .select_from(db.BatStats)
        .group_by(db.BatStats.player_id)
        .group_by(db.BatStats.player_team_id_bbref)
        .order_by(db.BatStats.player_id_mlb),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_bat_stats_by_team_for_player(cls, db_engine, mlb_id):
        s = select([cls]).where(cls.mlb_id == mlb_id)
        results = db_engine.execute(s).fetchall()
        return BatStatsMetrics.from_query_results(results) if results else []


class BatStats_By_Team_Year_View(db.Base):
    __table__ = create_view(
        name="batstats_by_team_year",
        selectable=select(
            [
                db.BatStats.player_id.label("id"),
                db.BatStats.player_id_mlb.label("mlb_id"),
                db.BatStats.player_id_bbref.label("bbref_id"),
                db.BatStats.season_id.label("season_id"),
                db.Season.year.label("year"),
                db.BatStats.player_team_id.label("player_team_id"),
                db.BatStats.player_team_id_bbref.label("player_team_id_bbref"),
                db.Assoc_Player_Team.stint_number.label("stint_number"),
                func.count(db.BatStats.id).label("total_games"),
                avg,
                obp,
                slg,
                ops,
                iso,
                bb_rate,
                k_rate,
                contact_rate,
                func.sum(db.BatStats.plate_appearances).label("plate_appearances"),
                func.sum(db.BatStats.at_bats).label("at_bats"),
                func.sum(db.BatStats.hits).label("hits"),
                func.sum(db.BatStats.runs_scored).label("runs_scored"),
                func.sum(db.BatStats.rbis).label("rbis"),
                func.sum(db.BatStats.bases_on_balls).label("bases_on_balls"),
                func.sum(db.BatStats.strikeouts).label("strikeouts"),
                func.sum(db.BatStats.doubles).label("doubles"),
                func.sum(db.BatStats.triples).label("triples"),
                func.sum(db.BatStats.homeruns).label("homeruns"),
                func.sum(db.BatStats.stolen_bases).label("stolen_bases"),
                func.sum(db.BatStats.caught_stealing).label("caught_stealing"),
                func.sum(db.BatStats.hit_by_pitch).label("hit_by_pitch"),
                func.sum(db.BatStats.intentional_bb).label("intentional_bb"),
                func.sum(db.BatStats.gdp).label("gdp"),
                func.sum(db.BatStats.sac_fly).label("sac_fly"),
                func.sum(db.BatStats.sac_hit).label("sac_hit"),
                func.sum(db.BatStats.total_pitches).label("total_pitches"),
                func.sum(db.BatStats.total_strikes).label("total_strikes"),
                func.sum(db.BatStats.wpa_bat).label("wpa_bat"),
                func.sum(db.BatStats.wpa_bat_pos).label("wpa_bat_pos"),
                func.sum(db.BatStats.wpa_bat_neg).label("wpa_bat_neg"),
                func.sum(db.BatStats.re24_bat).label("re24_bat"),
            ]
        )
        .select_from(
            join(
                join(db.BatStats, db.Season, db.BatStats.season_id == db.Season.id),
                db.Assoc_Player_Team,
                and_(
                    db.BatStats.player_id == db.Assoc_Player_Team.db_player_id,
                    db.BatStats.player_team_id == db.Assoc_Player_Team.db_team_id,
                ),
            )
        )
        .group_by(db.BatStats.season_id)
        .group_by(db.BatStats.player_id)
        .group_by(db.BatStats.player_team_id)
        .order_by(db.BatStats.player_id_mlb)
        .order_by(db.BatStats.season_id)
        .order_by(db.Assoc_Player_Team.stint_number),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_bat_stats_by_team_by_year_for_player(cls, db_engine, mlb_id):
        s = select([cls]).where(cls.mlb_id == mlb_id)
        results = db_engine.execute(s).fetchall()
        return BatStatsMetrics.from_query_results(results) if results else []


class BatStats_By_Opp_Team_View(db.Base):
    __table__ = create_view(
        name="batstats_by_opp_team",
        selectable=select(
            [
                db.BatStats.player_id.label("id"),
                db.BatStats.player_id_mlb.label("mlb_id"),
                db.BatStats.player_id_bbref.label("bbref_id"),
                db.BatStats.opponent_team_id_bbref.label("opponent_team_id_bbref"),
                func.count(db.BatStats.id).label("total_games"),
                avg,
                obp,
                slg,
                ops,
                iso,
                bb_rate,
                k_rate,
                contact_rate,
                func.sum(db.BatStats.plate_appearances).label("plate_appearances"),
                func.sum(db.BatStats.at_bats).label("at_bats"),
                func.sum(db.BatStats.hits).label("hits"),
                func.sum(db.BatStats.runs_scored).label("runs_scored"),
                func.sum(db.BatStats.rbis).label("rbis"),
                func.sum(db.BatStats.bases_on_balls).label("bases_on_balls"),
                func.sum(db.BatStats.strikeouts).label("strikeouts"),
                func.sum(db.BatStats.doubles).label("doubles"),
                func.sum(db.BatStats.triples).label("triples"),
                func.sum(db.BatStats.homeruns).label("homeruns"),
                func.sum(db.BatStats.stolen_bases).label("stolen_bases"),
                func.sum(db.BatStats.caught_stealing).label("caught_stealing"),
                func.sum(db.BatStats.hit_by_pitch).label("hit_by_pitch"),
                func.sum(db.BatStats.intentional_bb).label("intentional_bb"),
                func.sum(db.BatStats.gdp).label("gdp"),
                func.sum(db.BatStats.sac_fly).label("sac_fly"),
                func.sum(db.BatStats.sac_hit).label("sac_hit"),
                func.sum(db.BatStats.total_pitches).label("total_pitches"),
                func.sum(db.BatStats.total_strikes).label("total_strikes"),
                func.sum(db.BatStats.wpa_bat).label("wpa_bat"),
                func.sum(db.BatStats.wpa_bat_pos).label("wpa_bat_pos"),
                func.sum(db.BatStats.wpa_bat_neg).label("wpa_bat_neg"),
                func.sum(db.BatStats.re24_bat).label("re24_bat"),
            ]
        )
        .select_from(db.BatStats)
        .group_by(db.BatStats.player_id)
        .group_by(db.BatStats.opponent_team_id_bbref)
        .order_by(db.BatStats.player_id_mlb),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_bat_stats_by_opp_for_player(cls, db_engine, mlb_id):
        s = select([cls]).where(cls.mlb_id == mlb_id)
        results = db_engine.execute(s).fetchall()
        return BatStatsMetrics.from_query_results(results) if results else []


class BatStats_By_Opp_Team_Year_View(db.Base):
    __table__ = create_view(
        name="batstats_by_opp_team_year",
        selectable=select(
            [
                db.BatStats.player_id.label("id"),
                db.BatStats.player_id_mlb.label("mlb_id"),
                db.BatStats.player_id_bbref.label("bbref_id"),
                db.BatStats.season_id.label("season_id"),
                db.Season.year.label("year"),
                db.BatStats.opponent_team_id.label("opponent_team_id"),
                db.BatStats.opponent_team_id_bbref.label("opponent_team_id_bbref"),
                func.count(db.BatStats.id).label("total_games"),
                avg,
                obp,
                slg,
                ops,
                iso,
                bb_rate,
                k_rate,
                contact_rate,
                func.sum(db.BatStats.plate_appearances).label("plate_appearances"),
                func.sum(db.BatStats.at_bats).label("at_bats"),
                func.sum(db.BatStats.hits).label("hits"),
                func.sum(db.BatStats.runs_scored).label("runs_scored"),
                func.sum(db.BatStats.rbis).label("rbis"),
                func.sum(db.BatStats.bases_on_balls).label("bases_on_balls"),
                func.sum(db.BatStats.strikeouts).label("strikeouts"),
                func.sum(db.BatStats.doubles).label("doubles"),
                func.sum(db.BatStats.triples).label("triples"),
                func.sum(db.BatStats.homeruns).label("homeruns"),
                func.sum(db.BatStats.stolen_bases).label("stolen_bases"),
                func.sum(db.BatStats.caught_stealing).label("caught_stealing"),
                func.sum(db.BatStats.hit_by_pitch).label("hit_by_pitch"),
                func.sum(db.BatStats.intentional_bb).label("intentional_bb"),
                func.sum(db.BatStats.gdp).label("gdp"),
                func.sum(db.BatStats.sac_fly).label("sac_fly"),
                func.sum(db.BatStats.sac_hit).label("sac_hit"),
                func.sum(db.BatStats.total_pitches).label("total_pitches"),
                func.sum(db.BatStats.total_strikes).label("total_strikes"),
                func.sum(db.BatStats.wpa_bat).label("wpa_bat"),
                func.sum(db.BatStats.wpa_bat_pos).label("wpa_bat_pos"),
                func.sum(db.BatStats.wpa_bat_neg).label("wpa_bat_neg"),
                func.sum(db.BatStats.re24_bat).label("re24_bat"),
            ]
        )
        .select_from(join(db.BatStats, db.Season, db.BatStats.season_id == db.Season.id))
        .group_by(db.BatStats.season_id)
        .group_by(db.BatStats.player_id)
        .group_by(db.BatStats.opponent_team_id)
        .order_by(db.BatStats.player_id_mlb),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_bat_stats_by_opp_by_year_for_player(cls, db_engine, mlb_id):
        s = select([cls]).where(cls.mlb_id == mlb_id)
        results = db_engine.execute(s).fetchall()
        return BatStatsMetrics.from_query_results(results) if results else []



import shutil
import subprocess
from collections import defaultdict
from datetime import datetime

from getch import pause
from halo import Halo

import vigorish.database as db
from vigorish.cli.components import print_heading, print_message
from vigorish.constants import DATA_SET_TO_NAME_MAP, FAKE_SPINNER, JOB_SPINNER_COLORS
from vigorish.enums import DataSet, JobStatus, ScrapeTaskOption, StatusReport
from vigorish.scrape.bbref_boxscores.scrape_task import ScrapeBBRefBoxscores
from vigorish.scrape.bbref_games_for_date.scrape_task import ScrapeBBRefGamesForDate
from vigorish.scrape.brooks_games_for_date.scrape_task import ScrapeBrooksGamesForDate
from vigorish.scrape.brooks_pitch_logs.scrape_task import ScrapeBrooksPitchLogs
from vigorish.scrape.brooks_pitchfx.scrape_task import ScrapeBrooksPitchFx
from vigorish.status.report_status import report_date_range_status, report_season_status
from vigorish.util.datetime_util import get_date_range
from vigorish.util.dt_format_strings import DATE_ONLY_2
from vigorish.util.result import Result
from vigorish.util.sys_helpers import node_is_installed, node_modules_folder_exists

JOB_STATUS_TEXT_COLOR = {"scraped": "bright_green", "skipped": "blue"}
NODEJS_INSTALL_ERROR = (
    "Node.js installation cannot be located. Please install a recent, stable "
    "version and install dependencies from npm, see github repo for detailed "
    "intructions."
)
NPM_PACKAGES_INSTALL_ERROR = (
    "Nightmare is not installed, you must install it and other node dependencies in order to scrape any data."
)


class JobRunner:
    scrape_date: datetime
    data_set: DataSet

    def __init__(self, app, db_job):
        self.app = app
        self.config = app.config
        self.db_session = app.db_session
        self.scraped_data = app.scraped_data
        self.db_job = db_job
        self.job_id = self.db_job.id
        self.job_name = self.db_job.name
        self.data_sets = self.db_job.data_sets
        self.start_date = self.db_job.start_date
        self.end_date = self.db_job.end_date
        self.season = self.db_job.season
        self.spinners = defaultdict(lambda: Halo(spinner=FAKE_SPINNER))
        self.task_results = []
        self.scrape_task_option = self.app.get_current_setting("SCRAPE_TASK_OPTION")
        self.status_report = self.app.get_current_setting("STATUS_REPORT")

    def execute(self):
        return (
            self.initialize()
            .on_success(self.run_job)
            .on_success(self.show_status_report)
            .on_success(self.job_succeeded)
            .on_failure(self.job_failed)
        )

    def run_job(self):
        return (
            self.run_job_day_by_day()
            if self.scrape_task_option == ScrapeTaskOption.BY_DATE
            else self.scrape_date_range(self.start_date, self.end_date)
        )

    def run_job_day_by_day(self):
        for scrape_date in get_date_range(self.start_date, self.end_date):
            self.scrape_date = scrape_date
            result = self.scrape_date_range(scrape_date, scrape_date)
            if result.failure:
                return result
            self.task_results.clear()
        return Result.Ok()

    def scrape_date_range(self, start_date, end_date):
        for task_number, data_set in enumerate(self.data_sets, start=1):
            self.data_set = data_set
            self.report_task_results()
            self.update_spinner(data_set, task_number)
            result = self.scrape_data_set(data_set, start_date, end_date)
            if result.failure:
                if "skip" in result.error:
                    self.log_result_data_set_skipped(data_set, task_number)
                    self.spinners[data_set].stop()
                    continue
                return result
            self.log_result_data_set_scraped(data_set, task_number)
            self.spinners[data_set].stop()
        return Result.Ok()

    def scrape_data_set(self, data_set, start_date, end_date):
        task_dict = {
            DataSet.BROOKS_GAMES_FOR_DATE: ScrapeBrooksGamesForDate,
            DataSet.BROOKS_PITCH_LOGS: ScrapeBrooksPitchLogs,
            DataSet.BROOKS_PITCHFX: ScrapeBrooksPitchFx,
            DataSet.BBREF_GAMES_FOR_DATE: ScrapeBBRefGamesForDate,
            DataSet.BBREF_BOXSCORES: ScrapeBBRefBoxscores,
        }
        scrape_task = task_dict[data_set](self.app, self.db_job)
        return scrape_task.execute(start_date, end_date)

    def update_spinner(self, data_set, task_number):
        set_name = DATA_SET_TO_NAME_MAP[data_set]
        self.spinners[data_set].text = f"Scraping data set: {set_name} (Task {task_number}/{len(self.data_sets)})"
        self.spinners[data_set].color = JOB_SPINNER_COLORS[data_set]
        self.spinners[data_set].start()
        self.spinners[data_set].stop_and_persist(self.spinners[data_set].frame(), "")

    def log_result_data_set_scraped(self, data_set, task_number):
        set_name = DATA_SET_TO_NAME_MAP[data_set]
        scraped = f"Scraped data set: {set_name} (Task {task_number}/{len(self.data_sets)})"
        self.task_results.append((scraped, JOB_STATUS_TEXT_COLOR["scraped"]))

    def log_result_data_set_skipped(self, data_set, task_number):
        set_name = DATA_SET_TO_NAME_MAP[data_set]
        skipped = f"Skipped data set: {set_name} (Task {task_number}/{len(self.data_sets)})"
        self.task_results.append((skipped, JOB_STATUS_TEXT_COLOR["skipped"]))

    def report_task_results(self):
        subprocess.run(["clear"])
        job_name_id = f"Job Name: {self.job_name} (ID: {self.job_id.upper()})"
        if self.job_name == self.job_id:
            job_name_id = f"Job ID: {self.job_id.upper()}"
        start_date_str = self.start_date.strftime(DATE_ONLY_2)
        end_date_str = self.end_date.strftime(DATE_ONLY_2)
        date_range = f"Scraping: {start_date_str}-{end_date_str}"
        if self.scrape_date:
            scrape_date_str = self.scrape_date.strftime(DATE_ONLY_2)
            date_range += f" (Now: {scrape_date_str})"
        job_heading = f"{job_name_id} {date_range}"
        print_heading(job_heading, fg="bright_yellow")
        if not self.task_results:
            return
        for (message, text_color) in self.task_results:
            print_message(message, fg=text_color)
        return Result.Ok()

    def initialize(self):
        errors = []
        if not node_is_installed():
            errors.append(NODEJS_INSTALL_ERROR)
        elif not node_modules_folder_exists():
            errors.append(NPM_PACKAGES_INSTALL_ERROR)
        result = self.check_url_delay_settings()
        if result.failure:
            errors.append(result.error)
        result = self.check_s3_bucket() if self.s3_bucket_required() else Result.Ok()
        if result.failure:
            errors.append(result.error)
        result = self.create_all_folderpaths()
        if result.failure:
            errors.append(result.error)
        if not errors:
            self.start_time = datetime.now()
            return Result.Ok()
        return Result.Fail("\n".join(errors))

    def show_status_report(self):
        if self.status_report == StatusReport.NONE:
            pause(message="\nPress any key to continue...")
            return Result.Ok()
        result = (
            report_season_status(db_session=self.db_session, year=self.season.year, report_type=self.status_report)
            if self.status_report == StatusReport.SEASON_SUMMARY
            else report_date_range_status(
                db_session=self.db_session,
                start_date=self.db_job.start_date,
                end_date=self.db_job.end_date,
                report_type=self.status_report,
            )
        )
        if result.failure:
            return result
        report_viewer = result.value
        report_viewer.launch()
        return Result.Ok()

    def job_failed(self, error_message):
        self.end_time = datetime.now()
        self.db_job.status = JobStatus.ERROR
        new_error = db.ScrapeError(error_message=error_message, data_set=self.data_set, job_id=self.db_job.id)
        self.db_session.add(new_error)
        self.db_session.commit()
        return Result.Fail(error_message)

    def job_succeeded(self):
        self.end_time = datetime.now()
        self.db_job.status = JobStatus.COMPLETE
        self.db_session.commit()
        shutil.rmtree(self.db_job.scraped_html_root_folder)
        return Result.Ok()

    def create_all_folderpaths(self):
        return self.scraped_data.create_all_folderpaths(self.season.year)

    def check_url_delay_settings(self):
        return self.config.check_url_delay_settings(self.data_sets)

    def s3_bucket_required(self):
        return self.config.s3_bucket_required(self.data_sets)

    def check_s3_bucket(self):
        return self.scraped_data.check_s3_bucket()



"""Weather conditions at game start time and other descriptive info."""
from dataclasses import dataclass


@dataclass
class BBRefBoxscoreMeta:
    """Weather conditions at game start time and other descriptive info."""

    park_name: str = ""
    field_type: str = ""
    day_night: str = ""
    first_pitch_temperature: int = 0
    first_pitch_precipitation: str = ""
    first_pitch_wind: str = ""
    first_pitch_clouds: str = ""
    game_duration: str = ""
    attendance: int = 0

    def as_dict(self):
        """Convert boxscore meta info to a dictionary."""
        return {
            "__bbref_boxscore_meta__": True,
            "park_name": self.park_name,
            "field_type": self.field_type,
            "day_night": self.day_night,
            "first_pitch_temperature": self.first_pitch_temperature,
            "first_pitch_precipitation": self.first_pitch_precipitation,
            "first_pitch_wind": self.first_pitch_wind,
            "first_pitch_clouds": self.first_pitch_clouds,
            "game_duration": self.game_duration,
            "attendance": self.attendance,
        }



from dataclasses import dataclass


@dataclass
class UrlScrapeDelay:
    delay_is_required: bool
    delay_is_random: bool
    delay_uniform: int
    delay_random_min: int
    delay_random_max: int

    def __str__(self):
        if not self.delay_is_required:
            return "No delay after each URL"
        if not self.delay_is_random:
            return f"Delay is uniform ({self.delay_uniform} seconds)"
        return f"Delay is random ({self.delay_random_min}-{self.delay_random_max} seconds)"

    @property
    def delay_uniform_ms(self):
        return self.delay_uniform * 1000 if self.delay_uniform else 0

    @property
    def delay_random_min_ms(self):
        return self.delay_random_min * 1000 if self.delay_random_min else 0

    @property
    def delay_random_max_ms(self):
        return self.delay_random_max * 1000 if self.delay_random_max else 0

    @property
    def is_valid(self):
        if not self.delay_is_required:
            return False
        if not self.delay_is_random and self.delay_uniform < 3:
            return False
        return not self.delay_is_random or self.delay_random_min >= 3

    def to_dict(self):
        return {
            "URL_SCRAPE_DELAY_IS_REQUIRED": self.delay_is_required,
            "URL_SCRAPE_DELAY_IS_RANDOM": self.delay_is_random,
            "URL_SCRAPE_DELAY_IN_SECONDS": self.delay_uniform,
            "URL_SCRAPE_DELAY_IN_SECONDS_MIN": self.delay_random_min,
            "URL_SCRAPE_DELAY_IN_SECONDS_MAX": self.delay_random_max,
        }

    @staticmethod
    def from_config(config_dict):
        delay_is_required = config_dict.get("URL_SCRAPE_DELAY_IS_REQUIRED", True)
        delay_is_random = config_dict.get("URL_SCRAPE_DELAY_IS_RANDOM", True)
        delay_uniform = config_dict.get("URL_SCRAPE_DELAY_IN_SECONDS", 0)
        delay_random_min = config_dict.get("URL_SCRAPE_DELAY_IN_SECONDS_MIN", 3)
        delay_random_max = config_dict.get("URL_SCRAPE_DELAY_IN_SECONDS_MAX", 6)
        return UrlScrapeDelay(
            delay_is_required=delay_is_required,
            delay_is_random=delay_is_random,
            delay_uniform=delay_uniform,
            delay_random_min=delay_random_min,
            delay_random_max=delay_random_max,
        )

    @staticmethod
    def null_object():
        return {
            "URL_SCRAPE_DELAY_IS_REQUIRED": None,
            "URL_SCRAPE_DELAY_IS_RANDOM": None,
            "URL_SCRAPE_DELAY_IN_SECONDS": None,
            "URL_SCRAPE_DELAY_IN_SECONDS_MIN": None,
            "URL_SCRAPE_DELAY_IN_SECONDS_MAX": None,
        }
